/*
 * arch/arm/mach-integrator/hrtime.c
 *
 * High-Res Timer Implementation for EBSA285 based targets
 *
 * Author: George G. Davis <gdavis@mvista.com>
 *
 * Copyright 2003 (c) MontaVista, Software, Inc. 
 * 
 * This file is licensed under  the terms of the GNU General Public 
 * License version 2. This program is licensed "as is" without any 
 * warranty of any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/hrtime.h>
#include <linux/timex.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/errno.h>
#include <asm/hardware.h>
#include <asm/preem_latency.h>
#include <asm/mach/irq.h>
#include <asm/irq.h>


int
schedule_hr_timer_int(unsigned ref_jiffies, int ref_cycles)
{
	int temp_cycles;

	BUG_ON(ref_cycles < 0);

	/*
	 * Get offset from last jiffy
	 */
	temp_cycles = (ref_jiffies - jiffies) * arch_cycles_per_jiffy +
		ref_cycles - get_arch_cycles(jiffies);

	if(unlikely(temp_cycles <= 0))
		return -ETIME;

	BUG_ON(temp_cycles > 0xffffff);

	*CSR_TIMER2_LOAD = temp_cycles;
	*CSR_TIMER2_CNTL = TIMER_CNTL_ENABLE | TIMER_CNTL_DIV16;

	return 0;
}

int
get_arch_cycles(unsigned ref_jiffies)
{
	int ret;
	unsigned temp_jiffies;
	unsigned diff_jiffies;

	do {
		/* snapshot jiffies */
		temp_jiffies = jiffies;
		barrier();

		/* calculate cycles since the current jiffy */
		ret = HR_LATCH - *CSR_TIMER1_VALUE;

		/* compensate for ref_jiffies in the past */
		if (unlikely(diff_jiffies = jiffies - ref_jiffies))
			ret += diff_jiffies * arch_cycles_per_jiffy;

		barrier();
		/* repeat if we didn't have a consistent view of the world */
	} while(unlikely(temp_jiffies != jiffies));

	return ret;
}

static void
hr_timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	*CSR_TIMER2_CLR = 0;	/* clear interrupt */
	*CSR_TIMER2_CNTL = 0;	/* disable the timer */

	do_hr_timer_int();
}

static struct irqaction hr_timer_irq = {
	.name		= "high-res timer",
	.handler	= hr_timer_interrupt,
	.flags		= SA_INTERRUPT
};

static int 
hr_timer_init(void)
{
	*CSR_TIMER2_CLR = 0;
	*CSR_TIMER2_LOAD = 0;
	*CSR_TIMER2_CNTL = 0;

	return setup_arm_irq(IRQ_TIMER2, &hr_timer_irq);
}

__initcall(hr_timer_init);
