/*
 * arch/arm/mach-integrator/hrtime.c
 *
 * High-Res Timer Implementation for the ARM Integrator
 *
 * Author: George G. Davis <gdavis@mvista.com>
 *
 * Copyright 2003 (c) MontaVista, Software, Inc. 
 * 
 * This file is licensed under  the terms of the GNU General Public 
 * License version 2. This program is licensed "as is" without any 
 * warranty of any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/hrtime.h>
#include <linux/timex.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/errno.h>
#include <asm/hardware.h>
#include <asm/preem_latency.h>
#include <asm/mach/irq.h>
#include <asm/irq.h>


int
schedule_hr_timer_int(unsigned ref_jiffies, int ref_cycles)
{
	int temp_cycles;
	volatile TimerStruct_t * subhz_timer =
		(volatile TimerStruct_t *) TIMER2_VA_BASE;

	BUG_ON(ref_cycles < 0);

	/*
	 * Get offset from last jiffy
	 */
	temp_cycles = (ref_jiffies - jiffies) * arch_cycles_per_jiffy +
		ref_cycles - get_arch_cycles(jiffies);

	if(unlikely(temp_cycles <= 0))
		return -ETIME;

	BUG_ON(temp_cycles > 0xffff);

	subhz_timer->TimerLoad    = temp_cycles;
	subhz_timer->TimerControl = TIMER_CTRL;	/* enable timer */

	return 0;
}

int
get_arch_cycles(unsigned ref_jiffies)
{
	int ret;
	unsigned temp_jiffies;
	unsigned diff_jiffies;
	volatile TimerStruct_t * hz_timer =
		(volatile TimerStruct_t *) TIMER1_VA_BASE;

	do {
		/* snapshot jiffies */
		temp_jiffies = jiffies;
		barrier();

		/* calculate cycles since the current jiffy */
		ret = TIMER_RELOAD - (hz_timer->TimerValue & 0xffff);

		/* compensate for ref_jiffies in the past */
		if (unlikely(diff_jiffies = jiffies - ref_jiffies))
			ret += diff_jiffies * arch_cycles_per_jiffy;

		barrier();
		/* repeat if we didn't have a consistent view of the world */
	} while(unlikely(temp_jiffies != jiffies));

	return ret;
}

static void
hr_timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	volatile TimerStruct_t * subhz_timer =
		(volatile TimerStruct_t *) TIMER2_VA_BASE;

	subhz_timer->TimerClear = 1; /* clear interrupt */
	subhz_timer->TimerControl = 0; /* disable the timer */
	do_hr_timer_int();
}

static struct irqaction hr_timer_irq = {
	.name		= "high-res timer",
	.handler	= hr_timer_interrupt,
	.flags		= SA_INTERRUPT
};

static __init int 
hr_timer_init(void)
{
	int ret;
	volatile TimerStruct_t * subhz_timer =
		(volatile TimerStruct_t *) TIMER2_VA_BASE;

	subhz_timer->TimerControl = 0;	/* disabled */
	subhz_timer->TimerLoad    = 0;
	subhz_timer->TimerClear = 1;	/* clear pending interrupt */

	ret = setup_arm_irq(IRQ_TIMERINT2, &hr_timer_irq);

	return ret;
}

__initcall(hr_timer_init);
