/**************************************************************************
 *
 * arch/arm/mach-iop3xx/dma-copy.c
 *
 * copy_from_user and copy_to_user implementations using IOP321 DMA
 *
 * Authors: Dave Jiang (dave.jiang@intel.com)
 *          Carl Staelin (staelin@hpli69.hpli.hpl.hp.com)
 *          Deepak Saxena <dsaxena@mvista.com>
 *          David A. Griego (david.a.griego@intel.com)
 *
 * Maintainer: Deepak Saxena <dsaxena@mvista.com>
 *
 * Copyright (C) 2003 MontaVista Software, Inc.
 * Copyright (C) 2003 Intel Corporation
 * Copyright (C) 2003 Hewlett Packard
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This file implements the copy_from_user() and copy_to_user() APIs
 * using the IOP321's DMA engine. This code is NOT meant to be a generic
 * DMA API, but simply to implement the above two functions. Since the
 * 321 has 2 DMA engines, we use one for copy_to_user() and one for
 * copy_from_user(). We support both RAM<->RAM copies and RAM<->PCI copies.
 *
 *
 * History:    (02/13/2003, DJ) Initial Creation
 *             (02/27/2003, DJ) Integrated Carl's DMA code
 *             (04/17/2003, DS) Major cleanup, full testing of boundary cases
 *             (04/18/2003, DS) Added support for user PCI mappings
 *             (04/21/2003, DS) Interrupt driven support
 *             (06/03/2003, DJ) Back to polling due to locking issues
 *
 * TODO:
 *     - Too many spin locks, semaphores...head hurts...must clean
 *     - Discard "bogus" error state checking
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/iobuf.h>
#include <linux/mm.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/proc-fns.h>
#include <asm/proc/cache.h>
#include <asm/hardware.h>
#include <linux/irq.h>

/*
 * pick up local definitions
 */
#include "iop321-dma.h"
#include <asm/arch/iop321-dma.h>

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#include <linux/module.h>
#endif

#undef DEBUG
// #define DEBUG
#ifdef DEBUG
#define DPRINTK(s, args...) printk("321_DMA: " s, ## args)
#else
#define DPRINTK(s, args...)
#endif

/* globals */
iop321_dma_t dma_chan[2];	/* DMA channels */

/* static prototypes */
static int dma_copy_chaininit(iop321_dma_t * dma);
ssize_t iop321_dma_cpy(void *src, int size, void *dest, int wait);

static inline u32
page_offset(u32 virt)
{
	return virt & (PAGE_SIZE - 1);
}

static inline u32
page_remainder(u32 virt)
{
	return PAGE_SIZE - page_offset(virt);
}

static inline u32
page_base(u32 virt)
{
	return (virt & ~(PAGE_SIZE - 1));
}

static inline int
is_user(u32 virt)
{
	return (virt < PAGE_OFFSET);
}

static inline int
is_kernel_static(u32 virt)
{
	return ((virt >= PAGE_OFFSET) && (virt < (unsigned long) high_memory));
}

static inline int
is_kernel_vmalloc(u32 virt)
{
	return (virt >= VMALLOC_START && virt <= VMALLOC_END);
}

/*
 * map a kernel virtual address or kernel logical address to a phys address
 */
static inline u32
physical_address(u32 virt)
{
	pgd_t *pgdp;
	pmd_t *pmdp;
	pte_t *ptep;

	/* kernel static-mapped address */
	if (is_kernel_static(virt))
		return __pa((u32) virt);

	/* non-vmalloc so walk page tables */
	if (is_user(virt)) {
		pgdp = pgd_offset(current->mm, virt);
	} else {
		pgdp = pgd_offset_k(virt);
	}
	if (pgd_none(*pgdp) || pgd_bad(*pgdp))
		return 0;

	/* get pointer to page middle directory */
	pmdp = pmd_offset(pgdp, virt);
	if (pmd_none(*pmdp) || pmd_bad(*pmdp))
		return 0;

	/* get the PTE pointer */
	ptep = pte_offset(pmdp, virt);
	if (!ptep)
		return 0;

	return (page_base(pte_val(*ptep)) |
		page_offset(virt & (PAGE_SIZE - 1)));
}

/*
 * We don't support interrupts for now
 */
#if 0
static void
dma_copy_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	iop321_dma_t *dma = (iop321_dma_t *) dev_id;
	u32 csr = 0;
	sw_dma_t *sw_desc = NULL;

	csr = *(dma->regs.CSR);

	if (!(csr & (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK))) {
		return;
	}

	sw_desc = dma->last_desc->next;
	sw_desc->status = csr;
	sw_desc->complete = 1;

	spin_unlock(&sw_desc->process->mm->page_table_lock);
	wake_up_interruptible(&sw_desc->wait);

	*(dma->regs.CSR) |= csr;
}
#endif

int
dma_copy_init(void)
{
	int i;
	int ctr;
	sw_dma_t *sw_desc = NULL;
	void *desc = NULL;
	int err = 0;

	DPRINTK("Initializing...\n");

	for (ctr = 0; ctr < MAX_IOP321_DMA_CHANNEL; ctr++) {
		/* init free stack */
		INIT_LIST_HEAD(&dma_chan[ctr].free_q);
		INIT_LIST_HEAD(&dma_chan[ctr].process_q);
		/* init free stack spinlock */
		spin_lock_init(&dma_chan[ctr].free_lock);
		spin_lock_init(&dma_chan[ctr].process_lock);
		spin_lock_init(&dma_chan[ctr].chan_lock);
		sema_init(&dma_chan[ctr].chan_sem, 1);

		init_waitqueue_head(&dma_chan[ctr].wait);
		/* pre-alloc DMA descriptors */
		for (i = 0; i < MAX_DMA_DESC; i++) {
			/*
			 * we keep track of original address before alignment
			 * adjust so we can free it later
			 */
			sw_desc = kmalloc(sizeof (sw_dma_t), GFP_KERNEL);
			sw_desc->dma_virt = desc =
			    kmalloc((sizeof (dma_desc_t) + 0x20), GFP_KERNEL);
			memset(desc, 0, sizeof (dma_desc_t) + 0x20);

			/*
			 * hardware descriptors must be aligned on an
			 * 8-word boundary
			 */
			desc =
			    (dma_desc_t *) (((u32) desc & 0xffffffe0) + 0x20);

			/*
			 * HACK: There's not physaddr_map_single(),
			 * so we have to manually do an ioremap.
			 * 2.5 is supposed to fix this, but it fails
			 * miserably also. :)
			 */
			// get the physical address
			sw_desc->dma_phys = (u32) virt_to_phys(desc);
			// remap it to non-cached
			sw_desc->dma_desc =
			    (dma_desc_t *) ioremap(sw_desc->dma_phys,
						   sizeof (dma_desc_t));
			init_waitqueue_head(&sw_desc->wait);

			/* put the descriptors on the free stack */
			spin_lock_irq(&dma_chan[ctr].free_lock);
			list_add(&sw_desc->link, &dma_chan[ctr].free_q);
			spin_unlock_irq(&dma_chan[ctr].free_lock);
		}

	}

	// setting DMA channel 0
	dma_chan[0].regs.CCR = IOP321_DMA0_CCR;
	dma_chan[0].regs.CSR = IOP321_DMA0_CSR;
	dma_chan[0].regs.DAR = IOP321_DMA0_DAR;
	dma_chan[0].regs.NDAR = IOP321_DMA0_NDAR;
	dma_chan[0].regs.PADR = IOP321_DMA0_PADR;
	dma_chan[0].regs.PUADR = IOP321_DMA0_PUADR;
	dma_chan[0].regs.LADR = IOP321_DMA0_LADR;
	dma_chan[0].regs.BCR = IOP321_DMA0_BCR;
	dma_chan[0].regs.DCR = IOP321_DMA0_DCR;

	// setting DMA channel 1
	dma_chan[1].regs.CCR = IOP321_DMA1_CCR;
	dma_chan[1].regs.CSR = IOP321_DMA1_CSR;
	dma_chan[1].regs.DAR = IOP321_DMA1_DAR;
	dma_chan[1].regs.NDAR = IOP321_DMA1_NDAR;
	dma_chan[1].regs.PADR = IOP321_DMA1_PADR;
	dma_chan[1].regs.PUADR = IOP321_DMA1_PUADR;
	dma_chan[1].regs.LADR = IOP321_DMA1_LADR;
	dma_chan[1].regs.BCR = IOP321_DMA1_BCR;
	dma_chan[1].regs.DCR = IOP321_DMA1_DCR;

	err = dma_copy_chaininit(&dma_chan[0]);
	if (err < 0) {
		printk(KERN_ERR "Unable to setup DMA chain");
		return -EIO;
	}

	err = dma_copy_chaininit(&dma_chan[1]);
	if (err < 0) {
		printk(KERN_ERR "Unable to setup DMA chain");
		return -EIO;
	}

	DPRINTK("DMA init Done!\n");
	return 0;
}				/* end of dma_copy_init() */

void
dma_copy_exit(void)
{
	sw_dma_t *sw_desc = NULL;
	int ctr;

	for (ctr = 0; ctr < MAX_IOP321_DMA_CHANNEL; ctr++) {

		while (!list_empty(&dma_chan[ctr].free_q)) {
			sw_desc =
			    list_entry(dma_chan[ctr].free_q.next, sw_dma_t,
				       link);
			list_del(&sw_desc->link);
			iounmap(sw_desc->dma_desc);
			kfree(sw_desc->dma_virt);
			kfree((void *) sw_desc);
		}

		while (!list_empty(&dma_chan[ctr].process_q)) {
			sw_desc =
			    list_entry(dma_chan[ctr].process_q.next, sw_dma_t,
				       link);
			list_del(&sw_desc->link);
			iounmap(sw_desc->dma_desc);
			kfree(sw_desc->dma_virt);
			kfree((void *) sw_desc);
		}

		// get rid of the last dangle descriptor
		sw_desc = dma_chan[ctr].last_desc;
		iounmap(sw_desc->dma_desc);
		kfree(sw_desc->dma_virt);
		kfree((void *) sw_desc);

		free_irq(dma_chan[ctr].irq.EOC, (void *) &dma_chan[ctr]);
		free_irq(dma_chan[ctr].irq.EOT, (void *) &dma_chan[ctr]);
	}
}

static int
dma_copy_chaininit(iop321_dma_t * dma)
{
	int flags = 0;
	sw_dma_t *sw_desc = NULL;
	u32 csr = 0;

	if (down_interruptible(&dma->chan_sem)) {
		return 0;
	}

	spin_lock_irqsave(&dma->free_lock, flags);

	if (!list_empty(&dma->free_q)) {
		sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
		list_del(&sw_desc->link);
	} else {
		/*
		 * This should NEVER happen. We need to increase the amount
		 * of descriptors if that is the case
		 * XXX TODO: put in something nice later to handle this!
		 */
		spin_unlock_irqrestore(&dma->free_lock, flags);

		up(&dma->chan_sem);
		DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
		BUG();
		return -ENOMEM;
	}
	spin_unlock_irqrestore(&dma->free_lock, flags);

	DPRINTK("soft DMA desc acquired");

	// Setting byte count to 0 doesn't do anything except loading the
	// descriptor
	sw_desc->dma_desc->BC = 0;
	sw_desc->dma_desc->DC = DMA_DCR_MTM;
	sw_desc->dma_desc->NDAR = 0;
	sw_desc->next = NULL;

	spin_lock_irqsave(&dma->process_lock, flags);
	list_add(&sw_desc->link, &dma->process_q);
	spin_unlock_irqrestore(&dma->process_lock, flags);

	// Do DMA stuff
	*(dma->regs.CSR) |= (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK);
	*(dma->regs.NDAR) = sw_desc->dma_phys;
	*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE;

	DPRINTK("we are polling");
	csr = *(dma->regs.CSR);
	while (csr & DMA_CSR_CH_ACTIVE) {
		csr = *(dma->regs.CSR);
	}

	DPRINTK("removing from processing queue");
	spin_lock_irqsave(&dma->process_lock, flags);
	list_del(&sw_desc->link);
	spin_unlock_irqrestore(&dma->process_lock, flags);

	// check to see if DMA failed
	if (!(csr & DMA_CSR_ERR_MASK)) {
		dma->last_desc = sw_desc;
	} else {
		dma->last_desc = NULL;
		sw_desc->status = 0;
		sw_desc->complete = 0;
		spin_lock_irqsave(&dma->free_lock, flags);
		list_add(&sw_desc->link, &dma->free_q);
		spin_unlock_irqrestore(&dma->free_lock, flags);

		up(&dma->chan_sem);
		return -EIO;
	}

	DPRINTK("releasing sem");
	up(&dma->chan_sem);
	return 0;
}				/* end of dma_copy_chaininit() */

/*=======================================================================*/
/*  Procedure:  dma_copy_to_user()                                       */
/*                                                                       */
/*  Description:    DMA-based copy_to_user.                              */
/*                                                                       */
/*  Parameters:  to: destination address                                 */
/*               from: source address                                    */
/*               n: number of bytes to transfer                          */
/*                                                                       */
/*  Returns:     unsigned long: number of bytes NOT copied               */
/*                                                                       */
/*  Notes/Assumptions:                                                   */
/*              Assumes that kernel physical memory is contiguous, i.e., */
/*              the physical addresses of contiguous virtual addresses   */
/*              are also contiguous.                                     */
/*              Assumes that kernel memory doesn't get paged.            */
/*              Assumes that to/from memory regions cannot overlap       */
/*                                                                       */
/*  History:    Carl Staelin 1/27/03  Initial Creation                   */
/*              Dave Jiang   2/22/03  Attemped to DMA chaining           */
/*                                    (data corrupt with full chain)     */
/*                                    back to Carl's way for now         */
/*=======================================================================*/
unsigned long
dma_copy_to_user(void *to, const void *from, unsigned long n)
{
	u32 chunk;
	u32 from_chunk;
	u32 to_chunk;
	u32 to_pa = 0;
	iop321_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	int cflags = 0;
	u32 csr = 0;
	u32 unaligned_to, unaligned_from;
	void *u;
	const void *k;

	DPRINTK("dma_copy_to_user(%#10x, 0x%#10x, %lu): entering\n", (u32) to,
		(u32) from, n);
	to_pa = physical_address((u32) to);
	DPRINTK("  physaddr(to) = %#010x\n", to_pa);

	/*
	 * Walk the address range forcing page faults.  We must do writes
	 * instead of reads in case we have COW pages and need to get
	 * them re-mapped to a new address range.
	 *
	 * THIS IS SO UGLY I WANT TO CRY
	 */
	for (u = to, k = from; u < to + n; u += PAGE_SIZE, k += PAGE_SIZE)
		put_user(*(char *) k, (char *) u);

	/*
	 * We _only_ support mem -> mem and mem -> pci copies
	 */
	if (to_pa < PHYS_OFFSET &&
	    ((to_pa < IOP321_PCI_LOWER_MEM) || (to_pa > IOP321_PCI_UPPER_MEM)))
		return __arch_copy_to_user(to, from, n);

	/*
	 * Ok, start addr is not cache line-aligned, so we need to make it so.
	 */
	unaligned_to = (u32) to & 31;
	unaligned_from = (u32) from & 31;;
	if (unaligned_to | unaligned_from) {
		DPRINTK("Fixing up starting address\n");
		if (unaligned_from > unaligned_to) {
			DPRINTK("unaligned_from larger, copying %d bytes\n",
				32 - unaligned_to);
			__arch_copy_to_user(to, from, 32 - unaligned_to);
			to += (32 - unaligned_to);
			from += (32 - unaligned_to);
			n -= (32 - unaligned_to);
		} else {
			DPRINTK("unaligned_to larger, copying %d bytes\n",
				32 - unaligned_from);
			__arch_copy_to_user(to, from, 32 - unaligned_from);
			to += (32 - unaligned_from);
			from += (32 - unaligned_from);
			n -= (32 - unaligned_from);
		}
	}

	/*
	 * Ok, we're aligned at the top, now let's check the end
	 * of the buffer and align that. After this we should have
	 * a block that does not touch any extraneous cache lines.
	 */
	unaligned_to = ((u32) to + n) & 31;
	unaligned_from = ((u32) from + n) & 31;;
	if (unaligned_to | unaligned_from) {
		DPRINTK("Fixing ending alignment\n");
		if (unaligned_to > unaligned_from) {
			void *tmp_to = to + (n - unaligned_to);
			const void *tmp_from = from + (n - unaligned_to);

			DPRINTK("unaligned_to: Copying %d bytes\n",
				unaligned_to);

			__arch_copy_to_user(tmp_to, tmp_from, unaligned_to);

			n -= unaligned_to;
		} else {
			void *tmp_to = to + (n - unaligned_from);
			const void *tmp_from = from + (n - unaligned_from);

			DPRINTK("unaligned_from: Copying %d bytes\n",
				unaligned_from);

			__arch_copy_to_user(tmp_to, tmp_from, unaligned_from);

			n -= unaligned_from;
		}
	}

	/*
	 * OK! We should now be fully aligned on both ends.
	 */

	/*
	 * lock page tables
	 */
	spin_lock_irqsave(&current->mm->page_table_lock, cflags);

	dma = &dma_chan[0];

	if (down_interruptible(&dma->chan_sem)) {
		spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
		return n;
	}

	from_chunk = 0;
	to_chunk = 0;

	DPRINTK("USING DMA: %#010x, %#010x, %d\n", to, from, n);

	while (n > 0) {
		if (from_chunk == 0) {
			/* kernel logical address ==> contiguous */
			if (is_kernel_static((u32) from)) {
				from_chunk = n;
			} else {
				/* virtual address */
				from_chunk = page_remainder((u32) from);
			}
		}

		if (to_chunk == 0) {
			to_chunk = page_remainder((u32) to);
			to_pa = physical_address(page_base((u32) to));

			/*
			 * This should NOT happen...
			 */
			if (!to_pa) {
				DPRINTK
				    ("dma_copy_to_user: no physical address for 0x%x, falling back! **** \n",
				     (u32) to);
				up(&dma->chan_sem);
				spin_unlock_irqrestore(&current->mm->
						       page_table_lock, cflags);
				return __arch_copy_to_user(to, from, n);
			}
		}
		chunk = ((to_chunk < from_chunk) ? to_chunk : from_chunk);
		if (n < chunk) {
			chunk = n;
		}

		if (chunk == 0) {
			break;
		}

		/*
		 * Ensure that the cache is clean:
		 *      - from range must be cleaned
		 *      - to range must be invalidated
		 */
		cpu_dcache_clean_range((u32) from, (u32) from + chunk);
		cpu_dcache_invalidate_range((u32) to, (u32) to + chunk);

		/*
		 * TODO: Since we don't chain or anything fancy, we're
		 * don't need the overhead of the list. Just write directly
		 * to the register and then wait on the interrupt.
		 */
		if (!list_empty(&dma->free_q)) {
			sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// XXX TODO: put in something nice later to handle this!

			spin_unlock(&current->mm->page_table_lock);
			up(&dma->chan_sem);
			spin_unlock_irqrestore(&current->mm->page_table_lock,
					       cflags);
			DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
			BUG();
			return n;
		}

		sw_desc->dma_desc->PUADR = 0;
		sw_desc->dma_desc->NDAR = 0;

		/*  Is this a mem -> mem or mem -> PCI copy? */
		if (to_pa >= PHYS_OFFSET) {
			DPRINTK("   Mem->Mem\n");
			sw_desc->dma_desc->LADR = sw_desc->phys_dest =
			    to_pa + page_offset((u32) to);
			sw_desc->dma_desc->PADR = sw_desc->phys_src =
			    physical_address((u32) from);
			sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;
			sw_desc->dma_desc->DC = DMA_DCR_MTM | DMA_DCR_IE;
		} else {
			DPRINTK("   Mem->PCI\n");
			sw_desc->dma_desc->PADR = sw_desc->phys_dest =
			    to_pa + page_offset((u32) to);
			sw_desc->dma_desc->LADR = sw_desc->phys_src =
			    physical_address((u32) from);
			sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;
			sw_desc->dma_desc->DC = DMA_DCR_PCI_MWI | DMA_DCR_IE;
		}

		// Set last NDAR to NULL
		sw_desc->next = NULL;
		sw_desc->status = sw_desc->complete = 0;
		sw_desc->process = current;

		dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
		*(dma->regs.CCR) =
		    DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;

		from += chunk;
		to += chunk;
		n -= chunk;
		to_chunk -= chunk;
		from_chunk -= chunk;

#if 0
		// busy wait on channel until done
		csr = *(dma->regs.CSR);
		while (csr & DMA_CSR_CH_ACTIVE) {
			udelay(1);
			csr = *(dma->regs.CSR);
		}
#endif

		while ((csr = *(dma->regs.CSR)) & DMA_CSR_CH_ACTIVE) ;

		// put the last in the free queue
		dma->last_desc->status = 0;
		dma->last_desc->complete = 0;
		dma->last_desc->next = 0;
		list_add_tail(&dma->last_desc->link, &dma->free_q);

		// hookup the new last descriptor for chaining
		sw_desc->status = 0;
		sw_desc->complete = 0;
		dma->last_desc = sw_desc;

		// check to see if DMA failed...if so, we die
		// since something is VERY wrong.
		if (csr & DMA_CSR_ERR_MASK) {
			up(&dma->chan_sem);
			spin_unlock_irqrestore(&current->mm->page_table_lock,
					       cflags);
			BUG();
		}
	}

	up(&dma->chan_sem);
	spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);

	DPRINTK("dma_copy_to_user(0x%x, 0x%x, %lu): exiting\n", (u32) to,
		(u32) from, n);
	return n;
}

/*=======================================================================*/
/*  Procedure:  dma_copy_from_user()                                     */
/*                                                                       */
/*  Description:    DMA-based copy_from_user.                            */
/*                                                                       */
/*  Parameters:  to: destination address                                 */
/*               from: source address                                    */
/*               n: number of bytes to transfer                          */
/*                                                                       */
/*  Returns:     unsigned long: number of bytes NOT copied               */
/*                                                                       */
/*  Notes/Assumptions:                                                   */
/*              Assumes that kernel virtual memory is contiguous, i.e.,  */
/*              the physical addresses of contiguous virtual addresses   */
/*              are also contiguous.                                     */
/*              Assumes that kernel memory doesn't get paged.            */
/*              Assumes that to/from memory regions cannot overlap       */
/*              XXX this one doesn't quite work right yet                */
/*                                                                       */
/*  History:    Carl Staelin 1/27/03  Initial Creation                   */
/*              Dave Jiang   2/22/03  Attempted to DMA chaining          */
/*=======================================================================*/
unsigned long
dma_copy_from_user(void *to, const void *from, unsigned long n)
{
	u32 chunk;
	u32 from_chunk;
	u32 from_pa = 0;
	u32 to_chunk;
	iop321_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	int cflags = 0;
	u32 csr = 0;
	u32 unaligned_to, unaligned_from;
	const void *u;
	volatile char f;

	DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): entering\n", (u32) to,
		(u32) from, n);
	from_pa = physical_address((u32) to);
	DPRINTK("  physaddr(from) = %#010x\n", from_pa);

	/* Force paging of relevant pages */
	for (u = from; u < from + n; u += PAGE_SIZE)
		get_user(f, (char *) u);

	/*
	 * We _only_ support mem -> mem and mem -> pci copies
	 */
	if ((from_pa < PHYS_OFFSET) &&
	    ((from_pa < IOP321_PCI_LOWER_MEM)
	     || (from_pa > IOP321_PCI_UPPER_MEM)))
		return __arch_copy_from_user(to, from, n);
	/*
	 * Ok, start addr is not cache line-aligned, so we need to make it so.
	 */
	unaligned_to = (u32) to & 31;
	unaligned_from = (u32) from & 31;;
	if (unaligned_to | unaligned_from) {
		if (unaligned_from > unaligned_to) {
			__arch_copy_from_user(to, from, 32 - unaligned_to);
			to += 32 - unaligned_to;
			from += 32 - unaligned_to;
			n -= 32 - unaligned_to;
		} else {
			__arch_copy_from_user(to, from, 32 - unaligned_from);
			to += 32 - unaligned_from;
			from += 32 - unaligned_from;
			n -= 32 - unaligned_from;
		}
	}

	/*
	 * Ok, we're aligned at the top, now let's check the end
	 * of the buffer and align that. After this we should have
	 * a block that does not touch any extraneous cache lines.
	 */
	unaligned_to = ((u32) to + n) & 31;
	unaligned_from = ((u32) from + n) & 31;;
	if (unaligned_to | unaligned_from) {
		if (unaligned_to > unaligned_from) {
			void *tmp_to = to + (n - unaligned_to);
			const void *tmp_from = from + (n - unaligned_to);

			__arch_copy_from_user(tmp_to, tmp_from, unaligned_to);

			n -= unaligned_to;
		} else {
			void *tmp_to = to + n - unaligned_from;
			const void *tmp_from = from + n - unaligned_from;

			__arch_copy_from_user(tmp_to, tmp_from, unaligned_from);

			n -= unaligned_from;
		}
	}

	/*
	 * OK! We should now be fully aligned on both ends.
	 */
	/* Lock page table */
	spin_lock_irqsave(&current->mm->page_table_lock, cflags);

	dma = &dma_chan[1];

	if (down_interruptible(&dma->chan_sem)) {
		spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
		return n;
	}

	from_chunk = 0;
	to_chunk = 0;

	while (n > 0) {
		if (from_chunk == 0) {
			from_chunk = page_remainder((u32) from);
			from_pa = physical_address((u32) page_base((u32) from));
			if (!from_pa) {
				DPRINTK
				    ("dma_copy_from_user: no physical address for 0x%x,"
				     " falling back! **** \n", (u32) from);
				up(&dma->chan_sem);
				spin_unlock_irqrestore(&current->mm->
						       page_table_lock, cflags);
				return __arch_copy_from_user(to, from, n);
			}
		}

		if (to_chunk == 0) {
			/* kernel logical address ==> contiguous */
			if (is_kernel_static((u32) to)) {
				to_chunk = n;
			} else {
				/* kernel virtual address */
				to_chunk = page_remainder((u32) to);
			}
		}

		chunk = ((to_chunk < from_chunk) ? to_chunk : from_chunk);
		if (n < chunk) {
			chunk = n;
		}
		if (chunk == 0) {
			break;
		}

		/*
		 * Ensure that the cache is clean
		 */
		cpu_dcache_clean_range((u32) from, (u32) from + chunk);
		cpu_dcache_invalidate_range((u32) to, (u32) to + chunk);

		DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): to_chunk=%d, "
			"from_chunk=%d, chunk=%d\n", (u32) to, (u32) from,
			n, to_chunk, from_chunk, chunk);

		if (!list_empty(&dma->free_q)) {
			sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// XXX TODO: put in something nice later to handle this!
			up(&dma->chan_sem);
			spin_unlock_irqrestore(&current->mm->page_table_lock,
					       cflags);
			DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
			BUG();
			return n;
		}

		// Set last NDAR to NULL
		sw_desc->next = NULL;
		sw_desc->dma_desc->NDAR = 0;
		sw_desc->dma_desc->PUADR = 0;
		sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;

		/* Is this mem -> mem or mem -> PCI? */
		if (from_pa >= PHYS_OFFSET) {
			sw_desc->dma_desc->LADR = sw_desc->phys_dest =
			    physical_address((u32) to);
			sw_desc->dma_desc->PADR = sw_desc->phys_src =
			    from_pa + page_offset((u32) from);
			sw_desc->dma_desc->DC = DMA_DCR_MTM | DMA_DCR_IE;
		} else {
			sw_desc->dma_desc->LADR = sw_desc->phys_dest =
			    physical_address((u32) to);
			sw_desc->dma_desc->PADR = sw_desc->phys_src =
			    from_pa + page_offset((u32) from);
			sw_desc->dma_desc->DC = DMA_DCR_PCI_MRM | DMA_DCR_IE;
		}

		sw_desc->next = NULL;
		sw_desc->status = sw_desc->complete = 0;
		sw_desc->process = current;

		dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
		*(dma->regs.CCR) =
		    DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;

		from += chunk;
		to += chunk;
		n -= chunk;
		to_chunk -= chunk;
		from_chunk -= chunk;

#if 0
		// busy wait on channel until done
		csr = *(dma->regs.CSR);
		while (csr & DMA_CSR_CH_ACTIVE) {
			udelay(1);
			csr = *(dma->regs.CSR);
		}
#endif

		while ((csr = *(dma->regs.CSR)) & DMA_CSR_CH_ACTIVE) ;

		// put the last in the free queue
		dma->last_desc->status = 0;
		dma->last_desc->complete = 0;
		dma->last_desc->next = 0;
		list_add_tail(&dma->last_desc->link, &dma->free_q);

		// hookup the new last descriptor for chaining
		sw_desc->status = 0;
		sw_desc->complete = 0;
		dma->last_desc = sw_desc;

		/*
		 * If the DMA failed, the user has screwed up
		 * and he needs to debug his code.
		 */
		if (csr & DMA_CSR_ERR_MASK) {
			up(&dma->chan_sem);
			spin_unlock_irqrestore(&current->mm->page_table_lock,
					       cflags);
			BUG();
		}
	}

	up(&dma->chan_sem);
	spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);

	return n;
}

module_init(dma_copy_init);
module_exit(dma_copy_exit);
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(dma_copy_to_user);
EXPORT_SYMBOL(dma_copy_from_user);
