/*
 * Private Definitions for IOP321 DMA
 *
 * Author: Dave Jiang (dave.jiang@intel.com)
 * Copyright (C) 2002 Intel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#ifndef _321DMA_PRIVATE_H_
#define _321DMA_PRIVATE_H_

#define DEFAULT_DMA_IRQ_THRESH  10

#define DMA_DESC_SIZE 24

#define DMA_CCR_CLEAR               0x00000000
#define DMA_CCR_CHANNEL_ENABLE      0x00000001
#define DMA_CCR_CHAIN_RESUME        0x00000002

#define DMA_CSR_PCIX_SPLIT_ERROR    0x00000002
#define DMA_CSR_PCI_TGT_ABORT       0x00000004
#define DMA_CSR_PCI_MST_ABORT       0x00000008
#define DMA_CSR_INT_MST_ABORT       0x00000020
#define DMA_CSR_EOC_INT             0x00000100
#define DMA_CSR_EOT_INT             0x00000200
#define DMA_CSR_CH_ACTIVE           0x00000400
#define DMA_CSR_DONE_MASK           (DMA_CSR_EOC_INT | DMA_CSR_EOT_INT)

#define DMA_CSR_ERR_MASK            (DMA_CSR_PCIX_SPLIT_ERROR | DMA_CSR_PCI_TGT_ABORT | \
                                     DMA_CSR_PCI_MST_ABORT |  DMA_CSR_INT_MST_ABORT)

#define DMA_COMPLETE    0x1

/*
 * DMA Descriptor
 */
#pragma pack(1)
typedef struct _dma_desc {
	u32 NDAR;		/* next descriptor adress */
	u32 PADR;		/* PCI address */
	u32 PUADR;		/* upper PCI address */
	u32 LADR;		/* local address */
	u32 BC;			/* byte count */
	u32 DC;			/* descriptor control */
} dma_desc_t;
#pragma pack()

/*
 * DMA Software Descriptor
 */
typedef struct _sw_dma sw_dma_t;

struct _sw_dma {
	struct list_head link;
	dma_desc_t *dma_desc;	/* DMA descriptor pointer */
	void *dma_virt;		/* unaligned virtual dma addr */
	u32 dma_phys;		/* aligned DMA physical */
	u32 status;
	void *virt_src;		/* virt source */
	void *virt_dest;	/* virt destination */
	u32 phys_src;		/* physical/bus address of src */
	u32 phys_dest;		/* Physical/bus address of dest */
	u32 buf_size;
	sw_dma_t *next;
	wait_queue_head_t wait;
	int complete;
	struct task_struct *process;	/* Pointer to process that owns us */
};

/*
 * DMA control register structure
 */
#pragma pack(1)
typedef struct _dma_regs {
	volatile u32 *CCR;	/* channel control register */
	volatile u32 *CSR;	/* channel status register */
	volatile u32 *DAR;	/* descriptor address register */
	volatile u32 *NDAR;	/* next descriptor address register */
	volatile u32 *PADR;	/* PCI address register */
	volatile u32 *PUADR;	/* upper PCI address register */
	volatile u32 *LADR;	/* local address register */
	volatile u32 *BCR;	/* byte count register */
	volatile u32 *DCR;	/* descriptor control register */
} dma_regs_t;
#pragma pack()

/*
 * DMA channel structure.
 */

typedef struct _dmac_ctrl_t {
	spinlock_t chan_lock;
	struct semaphore chan_sem;
	struct list_head process_q;
	struct list_head free_q;
	spinlock_t process_lock;	/* process queue lock */
	spinlock_t free_lock;	/* hold queue lock */
	struct {
		int EOT;
		int EOC;
		int ERR;
	} irq;
	const char *device_id;	/* Device name */
	dma_regs_t regs;
	wait_queue_head_t wait;
	sw_dma_t *last_desc;
} iop321_dma_t;

#define SW_ENTRY(list) list_entry((list.next), sw_dma_t, link)

#endif
 /*EOF*/
