/*
 * arch/arm/mach-iop3xx/iop321-hrtime.c
 *
 * High-Res Timer Implementation for IOP321 boards
 *
 * Author: Deepak Saxena <dsaxena@mvista.com>
 *
 * Copyright 2003 (c) MontaVista, Software, Inc. 
 * 
 * This file is licensed under  the terms of the GNU General Public 
 * License version 2. This program is licensed "as is" without any 
 * warranty of any kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/hrtime.h>
#include <linux/timex.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/errno.h>
#include <asm/hardware.h>
#include <asm/preem_latency.h>
#include <asm/mach/irq.h>
#include <asm/irq.h>

unsigned int scaled_arch_cycles_per_nsec, 
		scaled_nsec_per_arch_cycle;

unsigned volatile last_jiffy_time;

#define USEC_PER_JIFFY (USEC_PER_SEC / HZ)

int
schedule_hr_timer_int(unsigned ref_jiffies, int ref_cycles)
{
	int temp_cycles;
	u32 timer_ctl = IOP321_TMR_EN | IOP321_TMR_PRIVILEGED | 
			IOP321_TMR_RATIO_1_1;

	BUG_ON(ref_cycles < 0);

	/*
	 * Get offset from last jiffy
	 */
	temp_cycles = (ref_jiffies - jiffies) * arch_cycles_per_jiffy + ref_cycles - get_arch_cycles(jiffies);
	if(unlikely(temp_cycles <= 0))
		return -ETIME;

	ref_cycles -= temp_cycles;
	asm volatile("mcr p6, 0, %0, c5, c1, 0" : : "r" (ref_cycles));

	asm volatile("mcr p6, 0, %0, c1, c1, 0"	: : "r" (timer_ctl));

	return 0;
}

int
get_arch_cycles(unsigned ref_jiffies)
{
	int ret;
	unsigned temp_jiffies;

	do {
		temp_jiffies = jiffies;
		barrier();

		if(unlikely(ref_jiffies > jiffies)) {
			ret = (ref_jiffies - jiffies) * arch_cycles_per_jiffy;
			ret += *IOP321_TU_TCR0;
			ret *= -1;
		} else {
			ret = arch_cycles_per_jiffy - *IOP321_TU_TCR0;
			if(unlikely(ref_jiffies < temp_jiffies))
				ret += (temp_jiffies - ref_jiffies) * 
						arch_cycles_per_jiffy;
		}

		barrier();	
	
	} while(unlikely(temp_jiffies != jiffies));

	return ret;
}

int
arch_cycle_to_nsec(int arch_cycles)
{
	int sign = arch_cycles < 1 ? -1 : 1;

	return sign * mpy_sc24(sign * arch_cycles, scaled_nsec_per_arch_cycle);
}

int
nsec_to_arch_cycle(int nsec)
{
	int sign = nsec < 1 ? -1 : 1;

	return sign * mpy_sc24(sign * nsec, scaled_arch_cycles_per_nsec);
}

static void
hr_time_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 tisr;

	asm volatile("mrc p6, 0, %0, c6, c1, 0" : "=r" (tisr));

	tisr |= 2;

	asm volatile("mcr p6, 0, %0, c6, c1, 0" : : "r" (tisr));

	do_hr_timer_int();
}

static struct irqaction hr_timer_irq = {
	name: "high-res timer",
	handler: hr_time_interrupt
};

void 
hr_time_init(void)
{
	scaled_arch_cycles_per_nsec = div_sc24(CLOCK_TICK_RATE, NSEC_PER_SEC);

	scaled_nsec_per_arch_cycle = div_sc24(NSEC_PER_SEC, CLOCK_TICK_RATE);

	setup_arm_irq(IRQ_IOP321_TIMER1, &hr_timer_irq);
}

