/*
 *  linux/arch/arm/kernel/dma.c
 *
 *  Copyright (C) 1995-2000 Russell King
 *  Copyright (C) 2003 MontaVista Software Inc <source@mvista.com>,
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Front-end to the DMA handling.  This handles the allocation/freeing
 *  of DMA channels, and provides a unified interface to the machines
 *  DMA facilities.
 */
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/mman.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>

#include <asm/dma.h>

#include <asm/arch/dma.h>
#include <asm/arch/hardware.h>
#include <asm/mach/dma.h>
#include <asm/mach/irq.h>

//#define debugprintk(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#define debugprintk(fmt, args...)

static struct irqaction dma_irq = {
	.name = "dma",
	.flags = SA_INTERRUPT,
};
static struct irqaction dma_err_irq = {
	.name = "dma_err",
	.flags = SA_INTERRUPT,
};

int
mx1ads_request_dma(dmach_t channel, dma_t * dma)
{
	_reg_DMA_DIMR |= (1 << channel);	/* disable interrupt for channel */

	dma->dma_callback = (callback_t) 0;
	dma->dma_err_callback = (err_callback_t) 0;

	return 0;
}

void
mx1ads_free_dma(dmach_t channel, dma_t * dma)
{
	_reg_DMA_DIMR |= (1 << channel);	/* disable interrupt for channel */

	dma->dma_callback = (callback_t) 0;
	dma->dma_err_callback = (err_callback_t) 0;
}

void
mx1ads_set_dma_addr(dmach_t channel, dma_t * dma, unsigned long physaddr)
{
	switch (dma->dma_mode & DMA_MODE_MASK) {
	case DMA_MODE_READ:
		*(((P_VU32) & _reg_DMA_DAR0) + channel * DMA_REG_SET_OFS) =
		    physaddr;
		debugprintk("mx1ads_set_dma_addr DAR %d, %d\n", channel,
			    physaddr);
		break;
	case DMA_MODE_WRITE:
		*(((P_VU32) & _reg_DMA_SAR0) + channel * DMA_REG_SET_OFS) =
		    physaddr;
		debugprintk("mx1ads_set_dma_addr SAR %d, %d\n", channel,
			    physaddr);
		break;
	default:		/* case DMA_MODE_CASCADE: */
		printk(KERN_ERR "dma: error setting dma addr for DMA%d\n",
		       channel);
		break;
	}
	return;
}

void
mx1ads_set_dma_count(dmach_t channel, dma_t * dma, unsigned long count)
{
	debugprintk("mx1ads_set_dma_count %d, %d\n", channel, count);
	*(((P_VU32) & _reg_DMA_CNTR0) + channel * DMA_REG_SET_OFS) = count;
}

void
mx1ads_enable_dma(dmach_t channel, dma_t * dma)
{
	debugprintk("mx1ads_enable_dma %d\n", channel);
	*(((P_VU32) & _reg_DMA_CCR0) + channel * DMA_REG_SET_OFS) |= 1;	/* set CEN bit */
	if (dma->dma_callback)
		_reg_DMA_DIMR &= ~((unsigned long) (1 << channel));	/* enable DMA intr */
}

void
mx1ads_disable_dma(dmach_t channel, dma_t * dma)
{
	debugprintk("mx1ads_disable_dma %d\n", channel);
	if (dma->dma_callback)
		_reg_DMA_DIMR |= (1 << channel);	/* disable DMA intr */
	*(((P_VU32) & _reg_DMA_CCR0) + channel * DMA_REG_SET_OFS) &= 0xFFFFFFFE;	/* clear CEN bit */
}

int
mx1ads_get_dma_residue(dmach_t channel, dma_t * dma)
{
	debugprintk("mx1ads_get_dma_residue %d, 0\n", channel);
	return 0;
}

extern dma_t *get_dma_channels(void);

void
mx1ads_request_dma_intr(dmach_t channel, callback_t dma_callback,
			err_callback_t dma_err_callback)
{
	dma_t *dma = get_dma_channels() + channel;

	if (channel >= MAX_DMA_CHANNELS)
		goto bad_dma;

	_reg_DMA_DIMR |= (1 << channel);	/* disable interrupt for channel */

	dma->dma_callback = dma_callback;
	dma->dma_err_callback = dma_err_callback;
	return;

      bad_dma:
	printk(KERN_ERR "dma: trying to request interrupt for DMA%d\n",
	       channel);
}

void
mx1ads_free_dma_intr(dmach_t channel)
{
	dma_t *dma = get_dma_channels() + channel;

	if (channel >= MAX_DMA_CHANNELS)
		goto bad_dma;

	_reg_DMA_DIMR |= (1 << channel);	/* disable interrupt for channel */

	dma->dma_callback = (callback_t) 0;
	dma->dma_err_callback = (err_callback_t) 0;

	return;

      bad_dma:
	printk(KERN_ERR "dma: trying to free interrupt for DMA%d\n", channel);
}

void
mx1ads_dma_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
	unsigned long bit_mask, intr_mask;
	dma_t *dma = get_dma_channels();

	debugprintk("!!! Now in DMA intr handler !!!\n");
	intr_mask = ~(_reg_DMA_DIMR) & 0x7FF;	/* we want '1' for enable */
	debugprintk("intr_mask: 0x%08x\n", intr_mask);
	debugprintk("DISR: 0x%08x\n", _reg_DMA_DISR);

	bit_mask = 1;
	/* scan through all the channels */
	for (i = 0; i < MAX_DMA_CHANNELS; i++) {
		if (intr_mask & _reg_DMA_DISR & bit_mask) {	/* if status set and enabled */
			debugprintk
			    ("Going to call callback function for channel %d\n",
			     i);
			debugprintk("dma->allocated: 0x%08x\n", dma->lock);
			debugprintk("dma->dma_callback: 0x%08x\n",
				    dma->dma_callback);
			if ((dma->lock) && (dma->dma_callback)) {
				debugprintk("Calling callback for channel %d\n",
					    i);
				(*((callback_t) (dma->dma_callback))) ();
				debugprintk("Back from callback.\n");
			}
			_reg_DMA_DISR |= bit_mask;	/* clear interrupt */
		}
		bit_mask <<= 1;
		dma++;
	}
	debugprintk("Exit DMA intr handler.\n");
}

void
mx1ads_dma_err_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;
	unsigned long bit_mask, intr_mask;
	dma_t *dma = get_dma_channels();

	debugprintk("!!! Now in DMA error intr handler !!!\n");
	intr_mask = ~(_reg_DMA_DIMR) & 0x7FF;	/* we want '1' for enable */
	bit_mask = 1;
	/* scan through all the channels */
	for (i = 0; i < MAX_DMA_CHANNELS; i++) {
		/* if burst time-out error bit set and interrupt enabled */
		if (intr_mask & _reg_DMA_DBTOSR & bit_mask) {
			if ((dma->lock) && (dma->dma_err_callback))
				(*(dma->dma_err_callback)) (MX1_DMA_ERR_BTO);
			_reg_DMA_DBTOSR |= bit_mask;
		}
		/* if request time-out error bit set and interrupt enabled */
		if (intr_mask & _reg_DMA_DRTOSR & bit_mask) {
			if ((dma->lock) && (dma->dma_err_callback))
				(*(dma->dma_err_callback)) (MX1_DMA_ERR_RTO);
			_reg_DMA_DRTOSR |= bit_mask;
		}
		/* if transfer error bit set and interrupt enabled */
		if (intr_mask & _reg_DMA_DSESR & bit_mask) {
			if ((dma->lock) && (dma->dma_err_callback))
				(*(dma->dma_err_callback)) (MX1_DMA_ERR_TE);
			_reg_DMA_DSESR |= bit_mask;
		}
		/* if buffer overflow error bit set and interrupt enabled */
		if (intr_mask & _reg_DMA_DBOSR & bit_mask) {
			if ((dma->lock) && (dma->dma_err_callback))
				(*(dma->dma_err_callback)) (MX1_DMA_ERR_BO);
		}
		bit_mask <<= 1;
		dma++;
	}
}

static struct dma_ops mx1ads_dops = {
	.request = mx1ads_request_dma,
	.free = mx1ads_free_dma,
	.enable = mx1ads_enable_dma,
	.disable = mx1ads_disable_dma,
	.residue = mx1ads_get_dma_residue,
	.setspeed = NULL,
	.type = "MX1ADS native DMA",
	.set_dma_count = mx1ads_set_dma_count,
	.set_dma_addr = mx1ads_set_dma_addr
};

void
arch_dma_init(dma_t * dma_chan)
{
	dma_t *dma;
	int i;
	debugprintk("mx1ads: arch_dma_init\n");

	/* initialize DMA module */
	_reg_DMA_DIMR = 0x7FF;	/* disable all interrupts */
	_reg_DMA_DBTOCR = 0;	/* disable burst time-out */

	/* initialize dma channel data structures */
	for (i = 0, dma = dma_chan; i < MAX_DMA_CHANNELS; i++, dma++) {
		dma->lock = 0;

		dma->dma_callback = (callback_t) 0;
		dma->dma_err_callback = (err_callback_t) 0;

		dma->d_ops = &mx1ads_dops;
	}

	/* register interrupt handler (don't use request_irq) */
	dma_irq.handler = mx1ads_dma_intr_handler;
	setup_arm_irq(61, &dma_irq);
	dma_err_irq.handler = mx1ads_dma_err_intr_handler;
	setup_arm_irq(60, &dma_err_irq);

	/* enable DMA module */
	_reg_DMA_DCR = 1;
}

/* New APIs for MX1 */
EXPORT_SYMBOL(mx1ads_request_dma_intr);
EXPORT_SYMBOL(mx1ads_free_dma_intr);
