/* Low-level polled-mode parallel port routines for the Attansic AT7601F
 *
 * Author: Ada Shangguann <Ada_Shangguan@sernet.com.cn>
 *
 * Inspiration taken from parport_amiga.c and parport_atari.c.
 *
 * To use, under menuconfig:
 *   1)  Turn on <*> Accelent IDP under Parallel port setup
 *   2)  Turn on <*> Parallel printer support under Character devices
 *
 * This will give you parport0 configured as /dev/lp0
 *
 * To make the correct /dev/lp* entries, enter /dev and type this:
 *
 * mknod lp0 c 6 0
 * mknod lp1 c 6 1
 * mknod lp2 c 6 2
 *
 */
 

#include <linux/module.h>
#include <linux/init.h>
#include <linux/parport.h>
#include <linux/interrupt.h>
#include <asm/hardware.h>
#include <asm/irq.h>

#define PORT_BASE		IXP425_EXP_BUS_CS6_BASE_VIRT
#define IRQ_SYSTEM		IRQ_IXP425_GPIO4

typedef struct
{
	volatile unsigned char data;
	volatile unsigned char status;
	volatile unsigned char control;
} AT7601F;

AT7601F *p_at7601;

/*
 * Parallel data port is port H, data
 * Parallel data direction is port H, direction
 * Control port is port I, data, lowest 4 bits
 * Status port is port G, data, upper 5 bits
 */


/* masks */
#define CONTROL_MASK		(PARPORT_CONTROL_STROBE 	|	\
							 PARPORT_CONTROL_AUTOFD		|	\
							 PARPORT_CONTROL_INIT		|	\
							 PARPORT_CONTROL_SELECT)
								 
#define STATUS_MASK			(PARPORT_STATUS_ERROR		|	\
							 PARPORT_STATUS_SELECT		|	\
							 PARPORT_STATUS_PAPEROUT	|	\
							 PARPORT_STATUS_ACK			|	\
							 PARPORT_STATUS_BUSY)


#define PARPORT_CONTROL_IRQ			(1<<4)		/* bit set == interrupt */
#define PARPORT_CONTROL_DIRECTION	(1<<5)		/* bit set == input */

#undef DEBUG

#ifdef DEBUG
#define DPRINTK  printk
#else
#define DPRINTK(stuff...)
#endif

static struct parport *this_port = NULL;

static unsigned char
parport_idp_read_data(struct parport *p)
{
	unsigned char c;

	c = p_at7601->data;
	DPRINTK("read_data:0x%x\n", c);
	return c;
}

static void
parport_idp_write_data(struct parport *p, unsigned char data)
{
	p_at7601->data = data;
	DPRINTK("write_data:0x%x\n", data);
}

static unsigned char
parport_idp_read_control(struct parport *p)
{
	unsigned char c;

	c = p_at7601->control & CONTROL_MASK;
	DPRINTK("read_control:0x%x\n", c);
	return c;
}

static void
parport_idp_write_control(struct parport *p, unsigned char control)
{
	unsigned int temp;

	temp = p_at7601->control;
	temp &= ~CONTROL_MASK;
	p_at7601->control = (temp | (control & CONTROL_MASK));
	DPRINTK("write_control:0x%x\n", control);
}

static unsigned char
parport_idp_frob_control(struct parport *p, unsigned char mask,
			   unsigned char val)
{
	unsigned char c;

	/* From the parport-lowlevel.txt file...*/
	/* This is equivalent to reading from the control register, masking out
	   the bits in mask, exclusive-or'ing with the bits in val, and writing
	   the result to the control register. */

	/* Easy enough, right? */

	c = parport_idp_read_control(p);
	parport_idp_write_control(p, (c & ~mask) ^ val);
	DPRINTK("frob_control:0x%x\n", c);
	return c;
}

static unsigned char
parport_idp_read_status(struct parport *p)
{
	unsigned char c;

	c = p_at7601->status & STATUS_MASK;
	c ^= PARPORT_STATUS_BUSY;  /* toggle S7 bit, active low */
	DPRINTK("read_status:0x%x\n", c);
	return c;
}

static void
parport_idp_init_state(struct pardevice *d, struct parport_state *s)
{
}

static void
parport_idp_save_state(struct parport *p, struct parport_state *s)
{
}

static void
parport_idp_restore_state(struct parport *p, struct parport_state *s)
{
}

static void
parport_idp_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	parport_generic_irq(irq, (struct parport *) dev_id, regs);
}

static void
parport_idp_enable_irq(struct parport *p)
{
	parport_idp_frob_control (p, PARPORT_CONTROL_IRQ, PARPORT_CONTROL_IRQ);
}

static void
parport_idp_disable_irq(struct parport *p)
{
	parport_idp_frob_control (p, PARPORT_CONTROL_IRQ, 0);
}

static void
parport_idp_data_forward(struct parport *p)
{
	parport_idp_frob_control (p, PARPORT_CONTROL_DIRECTION, 0x00);
	DPRINTK("data_forward:0x%x\n",0);
}

static void
parport_idp_data_reverse(struct parport *p)
{
	parport_idp_frob_control (p, PARPORT_CONTROL_DIRECTION, PARPORT_CONTROL_DIRECTION);
	DPRINTK("data_reverse:0x%x\n",0xff);
}


#ifdef DEBUG
static void testHardware(void)
{
	volatile unsigned char *p_reset = (unsigned char *) IXP425_EXP_BUS_CS5_BASE_VIRT;
	volatile unsigned char *p_ecp = (unsigned char *) IXP425_EXP_BUS_CS6_BASE_VIRT + 0x400;
	unsigned char c;
	
	DPRINTK("testHardware:p_reset = %p - p_at7601=%p\n", p_reset, p_at7601);

	/* reset the chip */
	*p_reset = 0x01;
	
	/* clear reset */
	*p_reset = 0x00;

#if 0
	/* enable to see if printer will print the number 3 */
	p_at7601->data = '3';
	p_at7601->control = PARPORT_CONTROL_STROBE;
#endif
	
	/* enter configuration mode */
	p_ecp[4] = 0x78;
	p_ecp[4] = 0x78;
	
	/* read some chip registers */
	
	p_ecp[4] = 0x20;
	c = p_ecp[5];
	DPRINTK("testHardware:chipID Register 1 = 0x%x\n", c);
	
	p_ecp[4] = 0x21;
	c = p_ecp[5];
	DPRINTK("testHardware:chipID Register 2 = 0x%x\n", c);

	p_ecp[4] = 0xF0;
	c = p_ecp[5];
	DPRINTK("testHardware:Mode Control Register = 0x%x\n", c);

	p_ecp[4] =  0xF1;
	c = p_ecp[5];
	DPRINTK("testHardware:Channel Select Register = 0x%x\n", c);

	p_ecp[4] =  0xF2;
	c = p_ecp[5];
	DPRINTK("testHardware:Chip Control Register = 0x%x\n", c);

	/* exit configuration mode */
	p_ecp[4] = 0xAA;
}
#endif

static void parport_idp_inc_use_count(void)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

static void parport_idp_dec_use_count(void)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static struct parport_operations parport_idp_ops = {
	parport_idp_write_data,
	parport_idp_read_data,

	parport_idp_write_control,
	parport_idp_read_control,
	parport_idp_frob_control,

	parport_idp_read_status,

	parport_idp_enable_irq,
	parport_idp_disable_irq,

	parport_idp_data_forward,
	parport_idp_data_reverse,

	parport_idp_init_state,
	parport_idp_save_state,
	parport_idp_restore_state,

	parport_idp_inc_use_count,
	parport_idp_dec_use_count,

	parport_ieee1284_epp_write_data,
	parport_ieee1284_epp_read_data,
	parport_ieee1284_epp_write_addr,
	parport_ieee1284_epp_read_addr,

	parport_ieee1284_ecp_write_data,
	parport_ieee1284_ecp_read_data,
	parport_ieee1284_ecp_write_addr,

	parport_ieee1284_write_compat,
	parport_ieee1284_read_nibble,
	parport_ieee1284_read_byte,
};



int parport_at7601_init(void)
{
	struct parport *p;
	
	if (this_port)
	{
		return 0;
	}

	
	/* memory map printer controller into local bus */
	/* chip select set to intel mode - byte access mode */
	
	*IXP425_EXP_CS5 = 0xbfff3c43;
	*IXP425_EXP_CS6 = 0xbfff3c43;
	p_at7601 = (AT7601F *) PORT_BASE;
	
	gpio_line_config(IXP425_GPIO_PIN_4, IXP425_GPIO_IN | IXP425_GPIO_FALLING_EDGE);
	gpio_line_isr_clear(IXP425_GPIO_PIN_4);

#ifdef DEBUG
	testHardware();
#endif
	
	p = parport_register_port(PORT_BASE, IRQ_SYSTEM, PARPORT_DMA_NONE, &parport_idp_ops);

	if (!p) return 0;  /* return 0 on failure */

	
	if (request_irq(IRQ_SYSTEM, parport_idp_interrupt, 0, p->name, p)) 
	{
			parport_unregister_port (p);
			return 0;
	}

	this_port=p;
	printk("%s: Attansic parallel port registered.\n", p->name);
	parport_proc_register(p);
	parport_announce_port(p);

	return 1;
}


#ifdef MODULE

MODULE_AUTHOR("Ada Shangguan");
MODULE_DESCRIPTION("Parport Driver for Attansic AT7601F");
MODULE_SUPPORTED_DEVICE("Attansic Parallel Port");
MODULE_LICENSE("GPL");

int
init_module(void)
{
	return parport_at7601_init() ? 0 : -ENODEV;
}

void
cleanup_module(void)
{
	free_irq (IRQ_SYSTEM, this_port);
	parport_proc_unregister(this_port);
	parport_unregister_port(this_port);
	this_port = NULL;
}
#endif

