/*
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <net/if_arp.h>
#include <string.h>
#include <syslog.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include "lan.h"

#define MAC_BCAST_ADDR		"\xff\xff\xff\xff\xff\xff"

struct packed_ether_header {
  u_int8_t  ether_dhost[ETH_ALEN];      /* destination eth addr */
  u_int8_t  ether_shost[ETH_ALEN];      /* source ether addr    */
  u_int16_t ether_type;                 /* packet type ID field */
} __attribute__((packed));

typedef struct arpMessage
{
  struct packed_ether_header	ethhdr;
  u_short htype;	/* hardware type (must be ARPHRD_ETHER) */
  u_short ptype;	/* protocol type (must be ETHERTYPE_IP) */
  u_char  hlen;		/* hardware address length (must be 6) */
  u_char  plen;		/* protocol address length (must be 4) */
  u_short operation;	/* ARP opcode */
  u_char  sHaddr[ETH_ALEN];	/* sender's hardware address */
  u_char  sInaddr[4];	/* sender's IP address */
  u_char  tHaddr[ETH_ALEN];	/* target's hardware address */
  u_char  tInaddr[4];	/* target's IP address */
  u_char  pad[18];	/* pad for min. Ethernet payload (60 bytes) */
} __attribute__((packed)) arpMessage;

const int inaddr_broadcast = INADDR_BROADCAST;

static int DebugFlag = 0;
unsigned char SenderHwAddr[ETH_ALEN];

/*****************************************************************************/
/*
 * arpCheck  -- This function will send ARP packet to request RequestProAddr's HW addr
 *              Return Value:
 *              1 ==> Success, so you can get HW addr from RequestHwAddr
 *              0 ==> Fail, I can not find RequestProAddr's HW
 *             -1 ==> Error, I can not send packet out!
 */
int arpCheck(char *IfName,unsigned char *RequestHwAddr, unsigned char *RequestProAddr){

  arpMessage ArpMsgSend,ArpMsgRecv;
  struct sockaddr addr;
  int j,i=0;
  int socket_fd;
  
  struct ifreq ifr;
  int sockd;
  
  if( openRawSocket(IfName, &socket_fd, ETH_P_ARP) != 0){
    syslog(LOG_ERR,"arpCheck: openRawSocket: %m\n");
    return -1;
  }


  /* Try to get HW address. */
  sockd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockd == -1) { /* Open socket error. */
	syslog(LOG_ERR,"arpCheck: Open socket err: %m\n");
	return -1;
  }
  memset(&ifr,0,sizeof(struct ifreq));
  strcpy(ifr.ifr_name,IfName);

  if(ioctl(sockd, SIOCGIFHWADDR, &ifr) < 0) {
    close(sockd);
    syslog(LOG_ERR,"arpCheck: Open socket err: %m\n");
	return -1;
  }
  memcpy(SenderHwAddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);


  memset(RequestHwAddr, 0, ETH_ALEN);
  
  /* Make ARP message. */
  memset(&ArpMsgSend,0,sizeof(arpMessage));
  memcpy(ArpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(ArpMsgSend.ethhdr.ether_shost,SenderHwAddr,ETH_ALEN);
  ArpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_ARP);

  ArpMsgSend.htype	= htons(ARPHRD_ETHER);
  ArpMsgSend.ptype	= htons(ETHERTYPE_IP);
  ArpMsgSend.hlen	= ETH_ALEN;
  ArpMsgSend.plen	= 4;
  ArpMsgSend.operation	= htons(ARPOP_REQUEST);
  memcpy(ArpMsgSend.sHaddr,SenderHwAddr,ETH_ALEN);
  memcpy(&ArpMsgSend.tInaddr,RequestProAddr,4);

  if ( DebugFlag ) syslog(LOG_DEBUG,
    "broadcasting ARPOP_REQUEST for %u.%u.%u.%u\n",
    ArpMsgSend.tInaddr[0],ArpMsgSend.tInaddr[1],
    ArpMsgSend.tInaddr[2],ArpMsgSend.tInaddr[3]);
  do
    {
      do
    	{
      	  if ( i++ > 4 ) return 0; /*  5 probes  */
      	  memset(&addr,0,sizeof(struct sockaddr));
      	  strcpy(addr.sa_data,IfName);
      	  if ( sendto(socket_fd,&ArpMsgSend,sizeof(arpMessage),0,
	   	&addr,sizeof(struct sockaddr)) == -1 )
	    {
	      syslog(LOG_ERR,"arpCheck: sendto: %m\n");
	      return -1;
	    }
    	}
      while ( peekfd(socket_fd,50000) ); /* 50 msec timeout */
      do
    	{
      	  memset(&ArpMsgRecv,0,sizeof(arpMessage));
      	  j=sizeof(struct sockaddr);
      	  if ( recvfrom(socket_fd,&ArpMsgRecv,sizeof(arpMessage),0,
		    (struct sockaddr *)&addr,&j) == -1 )
    	    {
      	      syslog(LOG_ERR,"arpCheck: recvfrom: %m\n");
      	      return -1;
    	    }
	  if ( ArpMsgRecv.ethhdr.ether_type != htons(ETHERTYPE_ARP) )
	    continue;
      	  if ( ArpMsgRecv.operation == htons(ARPOP_REPLY) )
	    {
	      if ( DebugFlag ) syslog(LOG_DEBUG,
	      "ARPOP_REPLY received from %u.%u.%u.%u for %u.%u.%u.%u\n",
	      ArpMsgRecv.sInaddr[0],ArpMsgRecv.sInaddr[1],
	      ArpMsgRecv.sInaddr[2],ArpMsgRecv.sInaddr[3],
	      ArpMsgRecv.tInaddr[0],ArpMsgRecv.tInaddr[1],
	      ArpMsgRecv.tInaddr[2],ArpMsgRecv.tInaddr[3]);
	    }
      	  else
	    continue;
      	  if ( memcmp(ArpMsgRecv.tHaddr,SenderHwAddr,ETH_ALEN) )
	    {
	      if ( DebugFlag )
	    	syslog(LOG_DEBUG,
	    	"target hardware address mismatch: %02X.%02X.%02X.%02X.%02X.%02X received, %02X.%02X.%02X.%02X.%02X.%02X expected\n",
	    	ArpMsgRecv.tHaddr[0],ArpMsgRecv.tHaddr[1],ArpMsgRecv.tHaddr[2],
	    	ArpMsgRecv.tHaddr[3],ArpMsgRecv.tHaddr[4],ArpMsgRecv.tHaddr[5],
	    	SenderHwAddr[0],SenderHwAddr[1],
	        SenderHwAddr[2],SenderHwAddr[3],
		    SenderHwAddr[4],SenderHwAddr[5]);
	      continue;
	    }
      	  if ( memcmp(&ArpMsgRecv.sInaddr,RequestProAddr,4) )
	    {
	      if ( DebugFlag )
	    	syslog(LOG_DEBUG,
	    	"sender IP address mismatch: %u.%u.%u.%u received, %u.%u.%u.%u expected\n",
	    	ArpMsgRecv.sInaddr[0],ArpMsgRecv.sInaddr[1],ArpMsgRecv.sInaddr[2],ArpMsgRecv.sInaddr[3],
	    	RequestProAddr[0],RequestProAddr[1],
	    	RequestProAddr[2],RequestProAddr[3]);
	      continue;
	    }
	    /* OK, Now we can get the HW addr. */
	    memcpy(RequestHwAddr, ArpMsgRecv.sHaddr, ETH_ALEN);
      	return 1;
    	}
      while ( peekfd(socket_fd,50000) == 0 );
    }
  while ( 1 );
  return 0;
}

int openRawSocket (char *IfName, int *s, u_short type) {
	int optval = 1;
    

	if((*s = socket (AF_INET, SOCK_PACKET, htons (type))) == -1) {
		syslog(LOG_DEBUG, "Could not open raw socket");
		return -1;
	}

	
	if(setsockopt (*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof (optval)) == -1) {
		syslog(LOG_DEBUG, "Could not setsocketopt on raw socket");
		return -1;
	}
	return 0;
}
