/*
 * files.c -- DHCP server file manipulation *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#ifdef _SNT_UDHCP_
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#endif /* _SNT_UDHCP_ */
#include <time.h>
#include <ctype.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"


/* on these functions, make sure you datatype matches */
static int read_ip(char *line, void *arg)
{
	struct in_addr *addr = arg;
	inet_aton(line, addr);
	return 1;
}


static int read_str(char *line, void *arg)
{
	char **dest = arg;
	int i;

	if (*dest) free(*dest);
	*dest = strdup(line);

	/* elimate trailing whitespace */
	for (i = strlen(*dest) - 1; i > 0 && isspace((*dest)[i]); i--);
	(*dest)[i > 0 ? i + 1 : 0] = '\0';
	return 1;
}


static int read_u32(char *line, void *arg)
{
	u_int32_t *dest = arg;
	*dest = strtoul(line, NULL, 0);
	return 1;
}


static int read_yn(char *line, void *arg)
{
	char *dest = arg;
	if (!strcasecmp("yes", line) || !strcmp("1", line) || !strcasecmp("true", line))
		*dest = 1;
	else if (!strcasecmp("no", line) || !strcmp("0", line) || !strcasecmp("false", line))
		*dest = 0;
	else return 0;

	return 1;
}


/* read a dhcp option and add it to opt_list */
static int read_opt(char *line, void *arg)
{
	struct option_set **opt_list = arg;
	char *opt, *val;
	char fail;
	struct dhcp_option *option = NULL;
	int length = 0;
	char buffer[255];
	u_int16_t result_u16;
	int16_t result_s16;
	u_int32_t result_u32;
	int32_t result_s32;

	int i;

	if (!(opt = strtok(line, " \t="))) return 0;

	for (i = 0; options[i].code; i++)
		if (!strcmp(options[i].name, opt)) {
			option = &(options[i]);
			break;
		}

	if (!option) return 0;

	do {
		val = strtok(NULL, ", \t");
		if (val) {
			fail = 0;
			length = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				read_ip(val, buffer);
				break;
			case OPTION_IP_PAIR:
				read_ip(val, buffer);
				if ((val = strtok(NULL, ", \t/-")))
					read_ip(val, buffer + 4);
				else fail = 1;
				break;
			case OPTION_STRING:
				length = strlen(val);
				if (length > 254) length = 254;
				memcpy(buffer, val, length);
				break;
			case OPTION_BOOLEAN:
				if (!read_yn(val, buffer)) fail = 1;
				break;
			case OPTION_U8:
				buffer[0] = strtoul(val, NULL, 0);
				break;
			case OPTION_U16:
				result_u16 = htons(strtoul(val, NULL, 0));
				memcpy(buffer, &result_u16, 2);
				break;
			case OPTION_S16:
				result_s16 = htons(strtol(val, NULL, 0));
				memcpy(buffer, &result_s16, 2);
				break;
			case OPTION_U32:
				result_u32 = htonl(strtoul(val, NULL, 0));
				memcpy(buffer, &result_u32, 4);
				break;
			case OPTION_S32:
				result_s32 = htonl(strtol(val, NULL, 0));
				memcpy(buffer, &result_s32, 4);
				break;
			default:
				break;
			}
			length += option_lengths[option->flags & TYPE_MASK];
			if (!fail)
				attach_option(opt_list, option, buffer, length);
		} else fail = 1;
	} while (!fail && option->flags & OPTION_LIST);
	return 1;
}


static struct config_keyword keywords[] = {
	/* keyword	handler   variable address		default */
	{"start",	read_ip,  &(server_config.start),	"192.168.0.20"},
	{"end",		read_ip,  &(server_config.end),		"192.168.0.254"},
	{"interface",	read_str, &(server_config.interface),	"ixp0"},
	{"option",	read_opt, &(server_config.options),	""},
	{"opt",		read_opt, &(server_config.options),	""},
	{"max_leases",	read_u32, &(server_config.max_leases),	"254"},
	{"remaining",	read_yn,  &(server_config.remaining),	"yes"},
#ifndef _SNT_UDHCP_
	{"auto_time",	read_u32, &(server_config.auto_time),	"7200"},
#else
	{"auto_time",	read_u32, &(server_config.auto_time),	"10"},
#endif /* _SNT_UDHCP_ */
	{"decline_time",read_u32, &(server_config.decline_time),"3600"},
	{"conflict_time",read_u32,&(server_config.conflict_time),"3600"},
	{"offer_time",	read_u32, &(server_config.offer_time),	"60"},
	{"min_lease",	read_u32, &(server_config.min_lease),	"60"},
#ifdef _SNT_UDHCP_
	{"lease_file",	read_str, &(server_config.lease_file),	"/var/udhcpd.leases"},
#else
	{"lease_file",	read_str, &(server_config.lease_file),	"/etc/udhcpd.leases"},
#endif /* _SNT_UDHCP_*/
	{"pidfile",	read_str, &(server_config.pidfile),	"/var/run/udhcpd.pid"},
	{"notify_file", read_str, &(server_config.notify_file),	""},
	{"siaddr",	read_ip,  &(server_config.siaddr),	"0.0.0.0"},
	{"sname",	read_str, &(server_config.sname),	""},
	{"boot_file",	read_str, &(server_config.boot_file),	""},
	{"",		NULL, 	  NULL,				""}
};


int read_config(char *file,struct host **U_host)
{
	FILE *in;
	char buffer[80], *token, *line;
	unsigned int chaddr[6];
	char addr[16];
	struct host *head=NULL,*p=NULL,*q=NULL;
	int i;

	for (i = 0; strlen(keywords[i].keyword); i++)
		if (strlen(keywords[i].def))
			keywords[i].handler(keywords[i].def, keywords[i].var);

	if (!(in = fopen(file, "r"))) {
	  LOG(LOG_ERR, "unable to open config file: %s", file);
	  printf("F<%s>L<%d>\n",__FILE__,__LINE__);
		return 0;
	}

	while (fgets(buffer, 80, in)) {
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';

		if(strstr(buffer,"host")==buffer){
			p=malloc(sizeof(struct host)+1);
			memset(p->hostname,0,16);
			sscanf(buffer,"%*s %s %*s",p->hostname);
		}
		if(strstr(buffer,"hardware ethernet")==buffer){
			sscanf(buffer,"%*s %*s %x:%x:%x:%x:%x:%x",&chaddr[0],&chaddr[1],&chaddr[2],&chaddr[3],&chaddr[4],&chaddr[5]);
			memset(p->hardware,0,16);
			for(i=0;i<6;i++)
				p->hardware[i]=chaddr[i];
		}
		if(strstr(buffer,"fixed-address")==buffer){
			sscanf(buffer,"%*s %s",addr);
			p->fixed_addr=inet_addr(addr);
			p->next=NULL;
			if(head==NULL)
			{
				head=p;
				q=head;
			}
			else
			{
				q->next=p;
				q=q->next;
			}
		}

		token = buffer + strspn(buffer, " \t");
		if (*token == '\0') continue;
		line = token + strcspn(token, " \t=");
		if (*line == '\0') continue;
		*line = '\0';
		line++;
		line = line + strspn(line, " \t=");
		if (*line == '\0') continue;

		for (i = 0; strlen(keywords[i].keyword); i++)
			if (!strcasecmp(token, keywords[i].keyword))
				keywords[i].handler(line, keywords[i].var);
	}
	fclose(in);
	*U_host=head;
	return 1;
}


/* the dummy var is here so this can be a signal handler */
void write_leases(int dummy)
{
	FILE *fp;
	unsigned int i;
	char buf[255];
	time_t curr = time(0);
	unsigned long lease_time;

	dummy = 0;

	if (!(fp = fopen(server_config.lease_file, "w"))) {
	   LOG(LOG_ERR, "Unable to open %s for writing", server_config.lease_file);
		return;
	}

	for (i = 0; i < server_config.max_leases; i++) {
		if (leases[i].yiaddr != 0) {
			if (server_config.remaining) {
				if (lease_expired(&(leases[i])))
					lease_time = 0;
				else lease_time = leases[i].expires - curr;
			} else lease_time = leases[i].expires;
			lease_time = htonl(lease_time);
			fwrite(leases[i].chaddr, 16, 1, fp);
			fwrite(&(leases[i].yiaddr), 4, 1, fp);
			fwrite(&lease_time, 4, 1, fp);
#ifdef _SNT_UDHCP_
			fwrite(leases[i].hostname,HOST_NAME_LEN,1,fp);
#endif /* _SNT_UDHCP_ */
		}
	}
	fclose(fp);

	if (server_config.notify_file) {
		sprintf(buf, "%s %s", server_config.notify_file, server_config.lease_file);
		system(buf);
	}
}

/* the dummy var is here so this can be a signal handler */
/*
 * del_lease    -- This function will delete the lease from the leases[i]
 *                 client MAC come from /var/udhcpd.delmac
 */
void del_lease(int dummy){
    FILE *fp;
    unsigned int ip;
    struct dhcpOfferedAddr *lease_p;

    fp = fopen(DEL_IP_FILE,"r");

    if(!fp){ /* Can not read file, return! */
        return;
    }
    if( fread(&ip,sizeof(int),1,fp) ){
        lease_p = find_lease_by_yiaddr(ip);
        if (lease_p){
            lease_p->yiaddr = 0;
            write_leases(0);
        }
    }
    fclose(fp);
    return;
}



void read_leases(char *file)
{
	FILE *fp;
	unsigned int i = 0,j;
	time_t curr = time(0);
	struct dhcpOfferedAddr lease;
#ifdef _SNT_UDHCP_
	int fd=-1;

	if(access("/var/udhcpd.leases",F_OK)<0)
		fd=creat("/var/udhcpd.leases",644);
	if(fd!=-1)
		close(fd);
#endif /* _SNT_UDHCP_ */
	if (!(fp = fopen(file, "r"))) {
	   LOG(LOG_ERR, "Unable to open %s for reading", file);
		return;
	}

	while (i < server_config.max_leases && (fread(&lease, sizeof lease, 1, fp) == 1)) {
		if (lease.yiaddr >= server_config.start && lease.yiaddr <= server_config.end) {
			leases[i].yiaddr = lease.yiaddr;
			leases[i].expires = ntohl(lease.expires);
#ifdef _SNT_UDHCP_
			for(j=0;j<16;j++)
				leases[i].hostname[j]=lease.hostname[j];
#endif /* _SNT_UDHCP_ */
			if (server_config.remaining) leases[i].expires += curr;
			memcpy(leases[i].chaddr, lease.chaddr, sizeof(lease.chaddr));
			i++;
		}
	}

	DEBUG(LOG_INFO, "Read %d leases", i);

	if (i == server_config.max_leases) {
		if (fgetc(fp) == EOF)
			/* might be helpfull to drop expired leases */
		   LOG(LOG_WARNING, "Too many leases while loading %s\n", file);
	}
	fclose(fp);
}
