#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <errno.h>
#include "dns.h"

#define __set_errno(val) errno=(val)

static const char	digits[] = "0123456789";
static int
printable(int ch) {
	return (ch > 0x20 && ch < 0x7f);
}
static int
special(int ch) {
	switch (ch) {
	case 0x22: /* '"' */
	case 0x2E: /* '.' */
	case 0x3B: /* ';' */
	case 0x5C: /* '\\' */
	/* Special modifiers in zone files. */
	case 0x40: /* '@' */
	case 0x24: /* '$' */
		return (1);
	default:
		return (0);
	}
}
int
ns_name_ntop(const u_char *src, char *dst, size_t dstsiz) {
	const u_char *cp;
	char *dn, *eom;
	u_char c;
	u_int n;

	cp = src;
	dn = dst;
	eom = dst + dstsiz;

	while ((n = *cp++) != 0) {
		if ((n & NS_CMPRSFLGS) != 0) {
			/* Some kind of compression pointer. */
			__set_errno (EMSGSIZE);
			return (-1);
		}
		if (dn != dst) {
			if (dn >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			*dn++ = '.';
		}
		if (dn + n >= eom) {
			__set_errno (EMSGSIZE);
			return (-1);
		}
		for ((void)NULL; n > 0; n--) {
			c = *cp++;
			if (special(c)) {
				if (dn + 1 >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = '\\';
				*dn++ = (char)c;
			} else if (!printable(c)) {
				if (dn + 3 >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = '\\';
				*dn++ = digits[c / 100];
				*dn++ = digits[(c % 100) / 10];
				*dn++ = digits[c % 10];
			} else {
				if (dn >= eom) {
					__set_errno (EMSGSIZE);
					return (-1);
				}
				*dn++ = (char)c;
			}
		}
	}
	if (dn == dst) {
		if (dn >= eom) {
			__set_errno (EMSGSIZE);
			return (-1);
		}
		*dn++ = '.';
	}
	if (dn >= eom) {
		__set_errno (EMSGSIZE);
		return (-1);
	}
	*dn++ = '\0';
	return (dn - dst);
}
int
ns_name_unpack(const u_char *msg, const u_char *eom, const u_char *src,
	       u_char *dst, size_t dstsiz)
{
	const u_char *srcp, *dstlim;
	u_char *dstp;
	int n, len, checked;

	len = -1;
	checked = 0;
	dstp = dst;
	srcp = src;
	dstlim = dst + dstsiz;
	if (srcp < msg || srcp >= eom) {
		__set_errno (EMSGSIZE);
		return (-1);
	}
	/* Fetch next label in domain name. */
	while ((n = *srcp++) != 0) {
		/* Check for indirection. */
		switch (n & NS_CMPRSFLGS) {
		case 0:
			/* Limit checks. */
			if (dstp + n + 1 >= dstlim || srcp + n >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			checked += n + 1;
			*dstp++ = n;
			memcpy(dstp, srcp, n);
			dstp += n;
			srcp += n;
			break;

		case NS_CMPRSFLGS:
			if (srcp >= eom) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			if (len < 0)
				len = srcp - src + 1;
			srcp = msg + (((n & 0x3f) << 8) | (*srcp & 0xff));
			if (srcp < msg || srcp >= eom) {  /* Out of range. */
				__set_errno (EMSGSIZE);
				return (-1);
			}
			checked += 2;
			/*
			 * Check for loops in the compressed name;
			 * if we've looked at the whole message,
			 * there must be a loop.
			 */
			if (checked >= eom - msg) {
				__set_errno (EMSGSIZE);
				return (-1);
			}
			break;

		default:
			__set_errno (EMSGSIZE);
			return (-1);			/* flag error */
		}
	}
	*dstp = '\0';
	if (len < 0)
		len = srcp - src;
	return (len);
}
int
ns_name_uncompress(const u_char *msg, const u_char *eom, const u_char *src,
		   char *dst, size_t dstsiz)
{
	u_char tmp[NS_MAXCDNAME];
	int n;
	
	if ((n = ns_name_unpack(msg, eom, src, tmp, sizeof tmp)) == -1)
		return (-1);
	if (ns_name_ntop(tmp, dst, dstsiz) == -1)
		return (-1);
	return (n);
}

int
my_dn_expand(const u_char *msg, const u_char *eom, const u_char *src,
	  char *dst, int dstsiz)
{
	int n = ns_name_uncompress(msg, eom, src, dst, (size_t)dstsiz);

	if (n > 0 && dst[0] == '.')
		dst[0] = '\0';
	return (n);
}

#ifdef dn_expand
#undef dn_expand
#endif
#define dn_expand my_dn_expand	

static union {
    HEADER1 hdr;
    unsigned char buf[PACKETSZ];
} response;

static char name[MAXDNAME];
static char *pStart, *responseend, *responsepos;
static int numanswers;
unsigned short pref;

void
alloc_free (char *x)
{
    free (x);
}

static unsigned short
getshort (unsigned char *c)
{
    unsigned short u;
    u = c[0];
    return (u << 8) + c[1];
}

static int
findmx (int wanttype)
{
    unsigned short rrtype;
    unsigned short rrdlen;
    int i;

    if (numanswers <= 0)
        return 2;
    --numanswers;
    if (responsepos == responseend)
        return DNS_SOFT;

    i = dn_expand (response.buf, responseend, responsepos, name, MAXDNAME);

    if (i < 0)
        return DNS_SOFT;
    responsepos += i;

    i = responseend - responsepos;
    if (i < 4 + 3 * 2)
        return DNS_SOFT;
    rrtype = getshort (responsepos);
    rrdlen = getshort (responsepos + 8);
    responsepos += 10;
    if (rrtype == wanttype) {
        if (rrdlen < 3)
            return DNS_SOFT;
        pref = (responsepos[0] << 8) + responsepos[1];
        if (dn_expand
            (response.buf, responseend, responsepos + 2, name, MAXDNAME) < 0)
            return DNS_SOFT;
        responsepos += rrdlen;
        return 1;
    }
    responsepos += rrdlen;
    return 0;
}

int
dns_mxip (char *host, struct MX **host_s, int *num)
{
    int alen;
    int nummx;
    int n, j, r;
    struct MX *mx;
    int an;

    res_init ();
    alen = res_query (host, C_IN, T_MX, response.buf, sizeof (response));
    if (alen <= 0)
        return -1;
    if (alen >= sizeof (response))
        alen = sizeof (response);
    pStart = (char *) response.buf + sizeof (HEADER);
    responseend = (char *) response.buf + alen;
    responsepos = pStart;
    n = ntohs (response.hdr.qn);
    while (n-- > 0) {
        j = dn_expand (response.buf, responseend, responsepos, name,
                       MAXDNAME);
        if (j < 0)
            return DNS_SOFT;
        responsepos += j;
        j = responseend - responsepos;
        if (j < QFIXEDSZ)       /* QFIXEDSZ = 4 */
            return DNS_SOFT;
        responsepos += QFIXEDSZ;
    }
    numanswers = ntohs (response.hdr.rrn);
    if (numanswers == 0)
        return DNS_SOFT;
    an = numanswers;
    mx = (struct MX *) malloc (numanswers * sizeof (struct MX));
    if (!mx)
        return DNS_MEM;
    nummx = 0;
    while ((r = findmx (T_MX)) != 2) {
        if (r == DNS_SOFT) {
            alloc_free ((char *) mx);
            return DNS_SOFT;
        }
        if (r == 1) {
            mx[nummx].p = pref;
            mx[nummx].name = 0;
            mx[nummx].name = (char *) malloc (strlen (name) + 1);
            if (!mx[nummx].name) {
                while (nummx > 0)
                    alloc_free (mx[--nummx].name);
                alloc_free ((char *) mx);
                return DNS_MEM;
            }
            strcpy (mx[nummx].name, name);
        }
        ++nummx;
    }
    *host_s = mx;
    *num = an;
    return 0;
}
