/*   ____  __  _ _____ ____     _ _            _
**  / ___||  \/ |_   _|  _ \___| (_) ___ _ __ | |_
**  \___ \| |\/| || | | |_)/ __| | |/ _ \ '_ \| __|
**   ___) | |  | || | |  _| (__| | |  __/ | | | |_
**  |____/|_|  |_||_| |_|  \___|_|_|\___|_| |_|\__|
**
**  SMTPclient -- simple SMTP client
**
**  This program is a minimal SMTP client that takes an email
**  message body and passes it on to a SMTP server (default is the
**  MTA on the local host). Since it is completely self-supporting,
**  it is especially suitable for use in restricted environments.
**
**  ======================================================================
**
**  Copyright (c) 1997 Ralf S. Engelschall, All rights reserved.
**
**  This program is free software; it may be redistributed and/or modified
**  only under the terms of either the Artistic License or the GNU General
**  Public License, which may be found in the SMTP source distribution.
**  Look at the file COPYING.
**
**  This program is distributed in the hope that it will be useful, but
**  WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  ======================================================================
**
**  smtpclient_main.c -- program source
**
**  Based on smtp.c as of August 11, 1995 from
**      W.Z. Venema,
**      Eindhoven University of Technology,
**      Department of Mathematics and Computer Science,
**      Den Dolech 2, P.O. Box 513, 5600 MB Eindhoven, The Netherlands.
*/
//#define _BB_DEBUG_
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include "dns.h"

#include "smtpclient_getopt.h"
#include "smtpclient_errno.h"
#include "smtpclient_vers.h"

static char *cc_addr = 0;
static char *err_addr = 0;
static char *from_addr = NULL;
char mailhost[64];
static int mailport = 25;
static char *reply_addr = 0;
static char *subject = 0;
static int mime_style = 0;
static int verbose = 0;
static char *msg_file = 0;
static int usesyslog = 0;
static int rflag = 0;

static FILE *sfp;
static FILE *rfp;
static FILE *msgfp;

#define dprintf  if (verbose) printf
#define dvprintf if (verbose) vprintf

/* hack for Ultrix */
#ifndef LOG_DAEMON
#define LOG_DAEMON 0
#endif
extern const char *resolve_name(const char *name, int mailbox);
int get_smtp_server(char *type, char *server, int *smtp_port);
int get_lan_ip(char ip[16]);

#ifdef _BB_DEBUG_
void bb_debug(char *buf);
char bb_buf[300];
#endif
/*
**  logging support
*/
void
log (char *str, ...)
{
    va_list ap;
    char buf[1024];

    va_start (ap, str);
    vsprintf (buf, str, ap);
    if (usesyslog)
        syslog (LOG_ERR, "SMTPclient: %s", buf);
    else
        fprintf (stderr, "SMTPclient: %s\n", buf);
    va_end (ap);
    return;
}

/*
**  usage page
*/
void
usage (void)
{
    fprintf (stderr, "Usage: smtp [options] recipients ...\n");
    fprintf (stderr, "\n");
    fprintf (stderr, "Message Header Options:\n");
    fprintf (stderr, "  -s, --subject=STR      subject line of message\n");
    fprintf (stderr, "  -f, --from=ADDR        address of the sender\n");
    fprintf (stderr,
             "  -r, --reply-to=ADDR    address of the sender for replies\n");
    fprintf (stderr,
             "  -e, --errors-to=ADDR   address to send delivery errors to\n");
    fprintf (stderr,
             "  -c, --carbon-copy=ADDR address to send copy of message to\n");
    fprintf (stderr, "\n");
    fprintf (stderr, "Processing Options:\n");
    fprintf (stderr,
             "  -S, --smtp-host=HOST   host where MTA can be contacted via SMTP\n");
    fprintf (stderr,
             "  -P, --smtp-port=NUM    port where MTA can be contacted via SMTP\n");
    fprintf (stderr,
             "  -M, --mime-encode      use MIME-style translation to quoted-printable\n");
    fprintf (stderr,
             "  -L, --use-syslog       log errors to syslog facility instead of stderr\n");
    fprintf (stderr, "\n");
    fprintf (stderr, "Giving Feedback:\n");
    fprintf (stderr,
             "  -v, --verbose          enable verbose logging messages\n");
    fprintf (stderr, "  -V, --version          display version string\n");
    fprintf (stderr, "  -h, --help             display this page\n");
    fprintf (stderr, "\n");
    return;
}

/*
**  version page
*/
void
version (void)
{
    fprintf (stdout, "%s\n", SMTPclient_Hello);
    fprintf (stdout, "\n");
    fprintf (stdout,
             "Copyright (c) 1997 Ralf S. Engelschall, All rights reserved.\n");
    fprintf (stdout, "\n");
    fprintf (stdout,
             "This program is distributed in the hope that it will be useful,\n");
    fprintf (stdout,
             "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    fprintf (stdout,
             "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    fprintf (stdout, "the GNU General Public License for more details.\n");
    fprintf (stdout, "\n");
    return;
}

/*
**  examine message from server
*/
int
get_response (void)
{
    char buf[BUFSIZ];

    while (fgets (buf, sizeof (buf), rfp)) {
        buf[strlen (buf) - 1] = 0;
        dprintf ("%s --> %s\n", mailhost, buf);
        if (!isdigit (buf[0]) || buf[0] > '3') {
//            log ("unexpected reply: %s", buf);
            return (1);
        }
        if (buf[4] != '-')
            break;
    }
    return 0;
}

/*
**  say something to server and check the response
*/
int
chat (char *fmt, ...)
{
    va_list ap;

    va_start (ap, fmt);
    vfprintf (sfp, fmt, ap);
    va_end (ap);

    va_start (ap, fmt);
    dprintf ("%s <-- ", mailhost);
    dvprintf (fmt, ap);
    va_end (ap);

    fflush (sfp);
    return (get_response ());
}

/*
**  transform to MIME-style quoted printable
**
**  Extracted from the METAMAIL version 2.7 source code (codes.c)
**  and modified to emit \r\n at line boundaries.
*/

static char basis_hex[] = "0123456789ABCDEF";

void
toqp (FILE * infile, FILE * outfile)
{
    int c;
    int ct = 0;
    int prevc = 255;

    while ((c = getc (infile)) != EOF) {
        if ((c < 32 && (c != '\n' && c != '\t'))
            || (c == '=') || (c >= 127) || (ct == 0 && c == '.')) {
            putc ('=', outfile);
            putc (basis_hex[c >> 4], outfile);
            putc (basis_hex[c & 0xF], outfile);
            ct += 3;
            prevc = 'A';        /* close enough */
        }
        else if (c == '\n') {
            if (prevc == ' ' || prevc == '\t') {
                putc ('=', outfile);    /* soft & hard lines */
                putc (c, outfile);
            }
            putc (c, outfile);
            ct = 0;
            prevc = c;
        }
        else {
            if (c == 'F' && prevc == '\n') {
                /*
                 * HORRIBLE but clever hack suggested by MTR for
                 * sendmail-avoidance
                 */
                c = getc (infile);
                if (c == 'r') {
                    c = getc (infile);
                    if (c == 'o') {
                        c = getc (infile);
                        if (c == 'm') {
                            c = getc (infile);
                            if (c == ' ') {
                                /* This is the case we are looking for */
                                fputs ("=46rom", outfile);
                                ct += 6;
                            }
                            else {
                                fputs ("From", outfile);
                                ct += 4;
                            }
                        }
                        else {
                            fputs ("Fro", outfile);
                            ct += 3;
                        }
                    }
                    else {
                        fputs ("Fr", outfile);
                        ct += 2;
                    }
                }
                else {
                    putc ('F', outfile);
                    ++ct;
                }
                ungetc (c, infile);
                prevc = 'x';    /* close enough -- printable */
            }
            else {
                putc (c, outfile);
                ++ct;
                prevc = c;
            }
        }
        if (ct > 72) {
            putc ('=', outfile);
            putc ('\r', outfile);
            putc ('\n', outfile);
            ct = 0;
            prevc = '\n';
        }
    }
    if (ct) {
        putc ('=', outfile);
        putc ('\r', outfile);
        putc ('\n', outfile);
    }
    return;
}


/*
**  main procedure
*/

struct option options[] = {
    {"subject", 1, NULL, 's'},
    {"from", 1, NULL, 'f'},
    {"replay-to", 1, NULL, 'r'},
    {"errors-to", 1, NULL, 'e'},
    {"carbon-copy", 1, NULL, 'c'},
    {"smtp-host", 1, NULL, 'h'},
    {"smtp-port", 1, NULL, 'p'},
    {"mime-encode", 0, NULL, 'M'},
    {"use-syslog", 0, NULL, 'L'},
    {"verbose", 0, NULL, 'v'},
    {"version", 0, NULL, 'V'},
    {"help", 0, NULL, 'h'}
};

int
main (int argc, char **argv)
{
    char buf[BUFSIZ];
    char my_name[BUFSIZ];
    char tmp_buf[256];
    struct sockaddr_in ssin;
    struct hostent *hp;
//    struct servent *sp;
    int c;
    int s;
    int r;
    int j;
    struct passwd *pwd;
    char *p;
//    char srcchar;
//    struct MX *host;
    int ret;
    char smtp_server[32];
    char smtp_type;
    int smtp_port;

    /*
     *  Go away when something gets stuck.
     */
    alarm (60);

    /*
     *  Parse options
     */
    while ((c =
            getopt_long (argc, argv, ":s:f:r:e:c:m:P:MLvVh", options,
                         NULL)) != EOF) {
        switch (c) {
        case 's':
            subject = optarg;
            break;
        case 'f':
            from_addr = optarg;
            break;
        case 'r':
            reply_addr = optarg;
            break;
        case 'e':
            err_addr = optarg;
            break;
        case 'c':
            cc_addr = optarg;
            break;
        case 'm':
            msg_file = optarg;
            break;
        case 'P':
            mailport = atoi (optarg);
            break;
        case 'M':
            mime_style = 1;
            break;
        case 'L':
            usesyslog = 1;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'V':
            version ();
            exit (0);
        case 'h':
            usage ();
            exit (0);
        default:
            fprintf (stderr, "SMTP: invalid option `%c'\n", optopt);
            fprintf (stderr, "Try `%s --help' for more information.\n",
                     argv[0]);
            exit (1);
        }
    }
    if (argc == optind) {
        fprintf (stderr, "SMTP: wrong number of arguments\n");
        fprintf (stderr, "Try `%s --help' for more information.\n", argv[0]);
        exit (1);
    }

    /*
     *  Open Syslog facility
     */
    if (usesyslog)
        openlog (argv[0], LOG_PID, LOG_DAEMON);
    /*
     *  Find out my own host name for HELO;
     *  if possible, get the FQDN.
     */
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "main<%d>:: \n", __LINE__);
    bb_debug(bb_buf);
#endif
    if (gethostname (my_name, sizeof (my_name) - 1) < 0) {
        log ("gethostname: %s", errorstr (errno));
        exit (1);
    }
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "main<%d>:: my_name<%s>\n", __LINE__, my_name);
    bb_debug(bb_buf);
#endif
    if ((hp = gethostbyname (my_name)) == NULL) {
        log ("%s: unknown host\n", my_name);
        exit (1);
    }
    strcpy (my_name, hp->h_name);
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "main<%d>:: my_name<%s>\n", __LINE__, my_name);
    bb_debug(bb_buf);
#endif

    /*
     *  Determine from address.
     */
    if (from_addr == NULL) {
        if ((pwd = getpwuid (getuid ())) == 0) {
            sprintf (buf, "userid-%d@%s", getuid (), my_name);
        }
        else {
        	char ip[16];
        	if(get_lan_ip(ip)==0)
        		sprintf (buf, "root@%s", ip);
        	else
            	sprintf (buf, "%s@%s", pwd->pw_name, my_name);
        }
        from_addr = strdup (buf);
    }

    /*
     *  Connect to smtp daemon on mailhost.
     */
    for (j = optind; j < argc; j++) {
        p = strchr (argv[j], '@');
        if (!p)
            continue;
        else
            sprintf (mailhost, "%s", p + 1);
/*        ret = dns_mxip (mailhost, &host, &num);
        if (ret)
            continue;
        for (i = 0; i < num; i++) {
            if ((hp = gethostbyname (host[i].name)) == NULL) {
                if (i == num - 1) {
                    log ("%s: unknown host\n", mailhost);
                    exit (1);
                }
                else
                    continue;
            }
            if (hp->h_addrtype != AF_INET) {
                log ("unknown address family: %d", hp->h_addrtype);
                exit (1);
            }*/
            {
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "main<%d>:: \n", __LINE__);
    bb_debug(bb_buf);
#endif
                if (get_smtp_server(&smtp_type, smtp_server, &smtp_port)!=0)
                    exit(1);
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "main<%d>:: type=<%d>, server<%s>,prot<%d>\n", __LINE__, smtp_type, smtp_server, smtp_port);
    bb_debug(bb_buf);
#endif
                if (smtp_type==0) /* domain */
                {
                    if ((hp = gethostbyname (smtp_server)) == NULL) {
                        log ("%s: unknown host\n", mailhost);
                        exit (1);
                    }
                    memset ((char *) &ssin, 0, sizeof (ssin));
                    memcpy ((char *) &ssin.sin_addr, hp->h_addr, hp->h_length);
                    ssin.sin_family = hp->h_addrtype;

                }
                else /* IP */
                {
                    memset ((char *) &ssin, 0, sizeof (ssin));
                    inet_aton(smtp_server,&(ssin.sin_addr));
                    ssin.sin_family = AF_INET;
                }
            }
/*			ret=inet_aton(mailhost,&(ssin.sin_addr));
			if(!ret){
				const char * addr = resolve_name(mailhost, 0);
				if (!addr) {
					(void)fprintf(stderr,
					    "ping: unknown host %s\n", mailhost);
					exit(1);
				}
				ssin.sin_addr.s_addr = inet_addr(addr);
			}
			ssin.sin_family = AF_INET;*/
            //ssin.sin_port = htons (mailport);
            ssin.sin_port = htons (smtp_port);
            if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
                log ("socket: %s", errorstr (errno));
                exit (1);
            }
            if (connect (s, (struct sockaddr *) &ssin, sizeof (ssin)) < 0) {
                log ("connect: %s", errorstr (errno));
                exit (1);
            }
            if ((r = dup (s)) < 0) {
                log ("dup: %s", errorstr (errno));
                exit (1);
            }
            if ((sfp = fdopen (s, "w")) == 0) {
                log ("fdopen: %s", errorstr (errno));
                exit (1);
            }
            if ((rfp = fdopen (r, "r")) == 0) {
                log ("fdopen: %s", errorstr (errno));
                exit (1);
            }
            /*
             *  Give out SMTP headers.
             */
            ret = get_response ();  /* banner */
            if (ret)
                continue;
            ret = chat ("HELO %s\r\n", my_name);
            if (ret)
                continue;
            ret = chat ("MAIL FROM: <%s>\r\n", from_addr);
            if (ret)
                continue;
            ret = chat ("RCPT TO: <%s>\r\n", argv[j]);
            if (ret)
                continue;
            ret = chat ("DATA\r\n");
            if (ret)
                continue;

            /*
             *  Give out Message header.
             */
            fprintf (sfp, "From: %s\r\n", from_addr);
            if (subject)
                fprintf (sfp, "Subject: %s\r\n", subject);

            if (reply_addr)
                fprintf (sfp, "Reply-To: %s\r\n", reply_addr);
            if (err_addr)
                fprintf (sfp, "Errors-To: %s\r\n", err_addr);
#if 0
            if ((pwd = getpwuid (getuid ())) == 0) {
                fprintf (sfp, "Sender: userid-%d@%s\r\n", getuid (), my_name);
            }
            else {
                fprintf (sfp, "Sender: %s@%s\r\n", pwd->pw_name, my_name);
            }
#endif
            fprintf (sfp, "To: %s", argv[j]);
            fprintf (sfp, "\r\n");
            if (cc_addr)
                fprintf (sfp, "Cc: %s\r\n", cc_addr);

            if (mime_style) {
                fprintf (sfp, "MIME-Version: 1.0\r\n");
                fprintf (sfp,
                         "Content-Type: text/plain; charset=ISO-8859-1\r\n");
                fprintf (sfp,
                         "Content-Transfer-Encoding: quoted-printable\r\n");
            }
            fprintf (sfp, "\r\n");
            if (rflag == 1)
                fprintf (sfp, "%s\r\n", tmp_buf);
            else{
                if(!msg_file)
                    while (fgets (buf, sizeof (buf), stdin)) {
                        if (rflag == 0)
                            sprintf (tmp_buf, "%s", buf);
                   		buf[strlen (buf) - 1] = 0;
                        if (strcmp (buf, ".") == 0)   /* quote alone dots */
                            fprintf (sfp, "..\r\n");
                    	else      /* pass thru mode */
                        	fprintf (sfp, "%s\r\n", buf);
                    }
                else{
            		if ((msgfp = fopen (msg_file, "r")) != 0)
                        while (fgets (buf, sizeof (buf), msgfp)) {
                            if (rflag == 0)
                                sprintf (tmp_buf, "%s", buf);
                            buf[strlen (buf) - 1] = 0;
                            if (strcmp (buf, ".") == 0)   /* quote alone dots */
                                fprintf (sfp, "..\r\n");
                            else      /* pass thru mode */
                                fprintf (sfp, "%s\r\n", buf);
                        }
                }
                fprintf (sfp, "\r\n");  /* terminate the text to tell MTA the end of messages */
                rflag = 1;
            }
            ret = chat (".\r\n");
            if (ret)
                continue;
            ret = chat ("QUIT\r\n");
            if (ret)
                continue;
            fclose (sfp);
            fclose (rfp);
	    break;
        }
//    }
    exit (0);
}

int get_smtp_server(char *type, char *server, int *smtp_port)
{
#define EMAIL_CONF "/var/system.conf"

	FILE *fp;
	char line[256], buf[2], srv[32], ip[16], port[6];
	char *p;

	fp=fopen(EMAIL_CONF,"r");
	if(!fp)
		return -1;

  while (fgets(line, 256, fp)) {
    if (strncmp(line, "smtp_type=", 10)==0)
    {
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: <%s>\n", __LINE__, line);
    bb_debug(bb_buf);
#endif
		line[strlen(line) - 1] = 0; /* kill '\n' */
		p = strchr(line, '=');
		if (p)
		{
            buf[0]=*(p+1); buf[1]='\0';
            *type=atoi(buf);
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: %d\n", __LINE__, *type);
    bb_debug(bb_buf);
#endif
        }
        else
            return -1;
    } else if (strncmp(line, "smtp_server=", 12)==0)
    {
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: <%s>\n", __LINE__, line);
    bb_debug(bb_buf);
#endif
		line[strlen(line) - 1] = 0; /* kill '\n' */
		p = strchr(line, '=');
		if (p)
		{
            strncpy(srv, p+1, 32);
            srv[31]='\0';
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: <%s>\n", __LINE__, srv);
    bb_debug(bb_buf);
#endif
        }
        else
            return -1;
    } else if (strncmp(line, "smtp_serverip=", 14)==0)
    {
		line[strlen(line) - 1] = 0; /* kill '\n' */
		p = strchr(line, '=');
		if (p)
		{
            strncpy(ip, p+1, 16);
            ip[15]='\0';
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: <%s>\n", __LINE__, ip);
    bb_debug(bb_buf);
#endif
        }
        else
            return -1;
    } else if (strncmp(line, "smtp_port=", 10)==0)
    {
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: <%s>\n", __LINE__, line);
    bb_debug(bb_buf);
#endif
		line[strlen(line) - 1] = 0; /* kill '\n' */
		p = strchr(line, '=');
		if (p)
		{
            strcpy(port, p+1);
            port[5]='\0';
            *smtp_port=atoi(port);
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: %d\n", __LINE__, *smtp_port);
    bb_debug(bb_buf);
#endif
        }
        else
            return -1;
    }
  }
    if (*type == 0)
        strcpy(server, srv);
    else strcpy(server, ip);
#ifdef _BB_DEBUG_
    sprintf(bb_buf, "get_srever<%d>:: %d, server<%s>\n", __LINE__, *type, server);
    bb_debug(bb_buf);
#endif

    fclose(fp);
    return 0;

}

int get_lan_ip(char ip[16])
{
	FILE *fp;
	char buf[128];

	fp=fopen("/var/system.conf","r");
	if(!fp)
		return -1;
	while(fgets(buf,128,fp)){
		if(strncmp("Lan_Ip=",buf,7)==0){
			if (strchr(buf, '\n'))
				*(strchr(buf,'\n'))=0;
			strcpy(ip,buf+7);
			break;
		}
	}
	fclose(fp);
	return 0;
}

#ifdef _BB_DEBUG_

void bb_debug(char *buf)
{
	FILE *fp;

	fp=fopen("/var/log/smtp.debug","a");
	if(!fp)
		return;
    fprintf(fp, buf);
    fclose(fp);
}
#endif
 /*EOF*/
